import logging
import mock

from ansible_galaxy.actions import install
from ansible_galaxy import exceptions
from ansible_galaxy import repository_spec
from ansible_galaxy import requirements
from ansible_galaxy.models.repository import Repository
from ansible_galaxy.models.repository_spec import RepositorySpec
from ansible_galaxy.models.requirement import Requirement, RequirementOps
from ansible_galaxy.models.requirement_spec import RequirementSpec

log = logging.getLogger(__name__)


def display_callback(msg, **kwargs):
    log.debug(msg)


def test_install_repos_empty_requirements(galaxy_context):
    requirements_to_install = []

    ret = install.install_repositories(galaxy_context,
                                       requirements_to_install=requirements_to_install,
                                       display_callback=display_callback)

    log.debug('ret: %s', ret)

    assert isinstance(ret, list)
    assert ret == []


def test_install_repositories(galaxy_context, mocker):
    repo_spec = RepositorySpec(namespace='some_namespace', name='some_name',
                               version='9.4.5')
    expected_repos = [Repository(repository_spec=repo_spec)]

    requirements_to_install = \
        requirements.from_requirement_spec_strings(['some_namespace.this_requires_some_name'])

    mocker.patch('ansible_galaxy.actions.install.install_repository',
                 return_value=expected_repos)

    ret = install.install_repositories(galaxy_context,
                                       requirements_to_install=requirements_to_install,
                                       display_callback=display_callback)

    log.debug('ret: %s', ret)

    assert isinstance(ret, list)
    assert ret == expected_repos


def test_install_repositories_no_deps_required(galaxy_context, mocker):
    needed_deps = []

    repository_specs_to_install = \
        [repository_spec.repository_spec_from_string('some_namespace.this_requires_nothing')]

    # mock out install_repository
    mocker.patch('ansible_galaxy.actions.install.install_repository',
                 return_value=[])

    ret = install.install_repositories(galaxy_context,
                                       requirements_to_install=repository_specs_to_install,
                                       display_callback=display_callback)

    log.debug('ret: %s', ret)

    assert isinstance(ret, list)
    assert ret == needed_deps


def test_verify_repository_specs_have_namespace_empty(galaxy_context):
    # will throw an exception if busted
    install._verify_requirements_repository_spec_have_namespaces([])


# even though 'blrp' isnt a valid spec, _build_content_set return something for now
def test_verify_repository_specs_have_namespace(galaxy_context):
    repository_spec = mock.Mock(requirement_spec=mock.Mock(namespace=None))
    try:
        install._verify_requirements_repository_spec_have_namespaces([repository_spec])
    except exceptions.GalaxyError as e:
        log.exception(e)
        return

    assert False, 'Expected a GalaxyError to be raised here since the repository_spec %s has no namespace or dots' % repository_spec


def test_find_new_deps_from_installed_no_deps(galaxy_context):
    res = install.find_new_deps_from_installed(galaxy_context, [], no_deps=True)
    assert res == []


def test_find_new_deps_from_installed_nothing_installed(galaxy_context):
    res = install.find_new_deps_from_installed(galaxy_context, [])
    assert res == []


def test_find_new_deps_from_installed(galaxy_context):
    repo_spec = RepositorySpec(namespace='some_namespace',
                               name='some_name',
                               version='4.3.2')

    req_spec = RequirementSpec(namespace='some_required_namespace',
                               name='some_required_name',
                               version_spec='==1.0.0')

    some_requirement = Requirement(repository_spec=repo_spec,
                                   op=RequirementOps.EQ,
                                   requirement_spec=req_spec)

    installed_repo = Repository(repo_spec, requirements=[some_requirement, some_requirement])
    res = install.find_new_deps_from_installed(galaxy_context, [installed_repo])

    log.debug('res: %s', res)
    assert isinstance(res, list)
    assert isinstance(res[0], Requirement)
    assert res[0].requirement_spec == req_spec



import logging
import os

from ansible_galaxy.actions import migrate_role
from ansible_galaxy import collection_info
from ansible_galaxy.models.collection_info import CollectionInfo
from ansible_galaxy.models.migrate_role_context import MigrateRoleContext

log = logging.getLogger(__name__)


OLD_SCHOOL_ROLE_PATH = os.path.join(os.path.dirname(__file__), '../', '../', 'data', 'roles', 'old_school_role')
COL_NAMESPACE = 'some_namespace'
COL_NAME = 'some_name'
COL_VERSION = '9.1.9'


def display_callback(msg, **kwargs):
    log.debug(msg)


def _migrate_role(output_path,
                  display_callback=display_callback,
                  role_path=OLD_SCHOOL_ROLE_PATH,
                  collection_namespace=COL_NAMESPACE,
                  collection_name=COL_NAME,
                  collection_version=COL_VERSION):

    migrate_role_context = MigrateRoleContext(role_path=role_path,
                                              output_path=output_path,
                                              collection_namespace=collection_namespace,
                                              collection_name=collection_name,
                                              collection_version=collection_version)
    res = migrate_role.migrate(migrate_role_context=migrate_role_context,
                               display_callback=display_callback)

    return res, migrate_role_context


def test_migrate_role_galaxy_yml(tmpdir):
    output_path = tmpdir.mkdir('mazer_test_migrate_role_action_test_migrate_role')

    res, migrate_role_context = _migrate_role(output_path=output_path.strpath)

    assert res == 0
    assert os.path.isdir(output_path.strpath)

    galaxy_yml_path = os.path.join(output_path.strpath, 'galaxy.yml')

    assert os.path.isfile(galaxy_yml_path)

    with open(galaxy_yml_path, 'r') as cfd:
        col_info = collection_info.load(cfd)

    log.debug('col_info: %s', col_info)

    assert isinstance(col_info, CollectionInfo)
    assert col_info.namespace == COL_NAMESPACE
    assert col_info.name == COL_NAME
    assert col_info.version == COL_VERSION
    assert isinstance(col_info.authors, list)
    assert isinstance(col_info.dependencies, dict)

# TODO: test missing role_path, output_path doesnt exist,
#       role loading errors, collection_info validation errors



import logging

import attr

from ansible_galaxy import repository
from ansible_galaxy import repository_archive

log = logging.getLogger(__name__)


def load_data_from_collection_artifact(repository_spec_string):
    log.debug('repo_spec_string: %s', repository_spec_string)

    # TODO: may need some name/path sanitations here
    archive_file_name = repository_spec_string

    repo_archive = repository_archive.load_archive(archive_file_name,
                                                   repository_spec=None)

    log.debug('repo_archive: %s', repo_archive)

    repo = repository.load_from_archive(repo_archive)

    # TODO: asdict?
    spec_data = repo.repository_spec
    log.debug('spec_data: %s', spec_data)

    # FIXME: we already have a valid RepositorySpec, but we dict'ify it here
    #        so existing code that assumes a dict continues to work
    return attr.asdict(spec_data)



import logging

from ansible_galaxy.models import migrate_role_context

log = logging.getLogger(__name__)


def test_init():
    role_path = "/dev/null/faux_role_path"
    output_path = "/dev/null/faux_output_path"
    mrc = migrate_role_context.MigrateRoleContext(role_path=role_path,
                                                  output_path=output_path)
    assert isinstance(mrc, migrate_role_context.MigrateRoleContext)
    assert mrc.role_path == role_path
    assert mrc.output_path == output_path


def test_repr():
    role_path = "/dev/null/faux_role_path"
    output_path = "/dev/null/faux_output_path"
    mrc = migrate_role_context.MigrateRoleContext(role_path=role_path,
                                                  output_path=output_path)

    r_mrc = repr(mrc)
    assert role_path in r_mrc
    assert output_path in r_mrc



import logging

import attr

from ansible_galaxy.models.repository_spec import RepositorySpec

log = logging.getLogger(__name__)


@attr.s(frozen=True)
class Repository(object):
    repository_spec = attr.ib(type=RepositorySpec)
    path = attr.ib(default=None)
    installed = attr.ib(default=False, type=bool, cmp=False)

    # ie, a role-as-collections-meta-main-deps
    dependencies = attr.ib(factory=tuple)

    # ie, a collection or role-as-collections-requirement.yml
    requirements = attr.ib(factory=tuple)

    # The data normally found in meta/main.yml for roles
    # meta_main = attr.ib(default=None, type=RoleMetadata)

    @property
    def label(self):
        return self.repository_spec.label

    def __str__(self):
        return '{repo_spec}@{path}'.format(repo_spec=self.repository_spec, path=self.path)



import datetime
import logging

import attr
import semantic_version

from ansible_galaxy.utils.version import convert_string_to_semver

log = logging.getLogger(__name__)


@attr.s(frozen=True)
class InstallInfo(object):
    '''The info that is saved into the .galaxy_install_info file'''
    install_date = attr.ib()
    install_date_iso = attr.ib(type=datetime.datetime)

    version = attr.ib(type=semantic_version.Version, default=None,
                      converter=convert_string_to_semver)

    @classmethod
    def from_version_date(cls, version, install_datetime):
        inst = cls(version=version,
                   install_date_iso=install_datetime,
                   install_date=install_datetime.strftime('%c'))
        return inst

    def to_dict_version_strings(self):
        data = attr.asdict(self)

        if data.get('verison', '') is None:
            del data['version']

        # semantic_version.Version isnt yaml-able, so build a dict with
        # the Version replaced with a str version
        ver = data.get('version', '')
        if ver:
            data['version'] = str(ver)

        return data



import datetime
import logging

import attr
import pytest

from ansible_galaxy.models import install_info

log = logging.getLogger(__name__)


def test_init():
    install_date = datetime.datetime.utcnow()
    info = install_info.InstallInfo(version='1.2.3',
                                    install_date=install_date.strftime('%c'),
                                    install_date_iso=install_date)

    log.debug('info: %s', info)

    assert isinstance(info, install_info.InstallInfo)
    assert isinstance(attr.asdict(info), dict)


def test_from_version_date():
    install_datetime = datetime.datetime.utcnow()
    info = install_info.InstallInfo.from_version_date(version='1.2.3',
                                                      install_datetime=install_datetime)

    log.debug('info: %s', info)

    assert isinstance(info, install_info.InstallInfo)
    assert isinstance(attr.asdict(info), dict)

    assert info.install_date_iso == install_datetime


def test_frozen():
    install_date = datetime.datetime.utcnow()
    info = install_info.InstallInfo(version='1.2.3',
                                    install_date=install_date.strftime('%c'),
                                    install_date_iso=install_date)

    log.debug('info: %s', info)

    with pytest.raises(attr.exceptions.FrozenInstanceError):
        info.version = '2.3.4'

    with pytest.raises(attr.exceptions.FrozenInstanceError):
        info.install_data_iso = datetime.datetime.utcnow()

    with pytest.raises(attr.exceptions.FrozenInstanceError):
        info.install_data_iso = datetime.datetime.utcnow().strftime('%c')


def test_equal():
    install_date1 = datetime.datetime.utcnow()
    info1 = install_info.InstallInfo(version='1.2.3',
                                     install_date=install_date1.strftime('%c'),
                                     install_date_iso=install_date1)

    info1a = install_info.InstallInfo(version='1.2.3',
                                      install_date=install_date1.strftime('%c'),
                                      install_date_iso=install_date1)

    info2 = install_info.InstallInfo(version='9.9.9',
                                     install_date=install_date1.strftime('%c'),
                                     install_date_iso=install_date1)

    assert info1 == info1a
    assert info1a == info1

    assert not info1 != info1a
    assert not info1a != info1

    assert not info2 == info1
    assert not info1 == info2



import logging
import pytest

from ansible_galaxy.models.collection_info import CollectionInfo

log = logging.getLogger(__name__)

NON_ALPHA_ERROR_PAT = r"Invalid collection metadata. Expecting 'name' and 'namespace' to contain only lowercase alphanumeric characters "
"or '_' only but '.*' contains others"


@pytest.fixture
def col_info():
    test_data = {
        'namespace': 'foo',
        'name': 'foo',
        'authors': ['chouseknecht'],
        'license': 'GPL-3.0-or-later',
        'version': '0.0.1',
        'description': 'unit testing thing',
    }
    return test_data


valid_name_or_namespaces = \
    ['some_name',
     'some_namespace',
     'roles2go',
     'ews',
     'ews_',
     'geerlingguy',
     'red_hat',
     ]


@pytest.fixture(scope='module',
                params=valid_name_or_namespaces)
def valid_name(request):
    yield request.param


def test_valid_names(col_info, valid_name):
    col_info['name'] = valid_name

    log.debug('valid_name: %s', valid_name)

    res = CollectionInfo(**col_info)

    assert isinstance(res, CollectionInfo)
    assert res.name == valid_name


invalid_name_or_namespaces = \
    [
     '2punk__rock--Valid|n a M e s',
     '__builtin__',
     '../../../bin/bash',
     r'\.',
     r'C:\huh\what',
     r'C:\AUX',
     ]


@pytest.fixture(scope='module',
                params=invalid_name_or_namespaces)
def invalid_name(request):
    yield request.param


def test_invalid_names(col_info, invalid_name):
    col_info['name'] = invalid_name

    log.debug('invalid_name: %s', invalid_name)

    with pytest.raises(ValueError) as exc:
        CollectionInfo(**col_info)

    log.debug('exc: %s', str(exc))


invalid_name_leading_underscore = \
    [
        '_scoop_co',
        '__dunder_mifflin',
        '_',
        '_foo',
    ]


@pytest.fixture(scope='module',
                params=invalid_name_leading_underscore)
def invalid_name_leading_underscore(request):
    yield request.param


def test_name_invalid_names_leading_underscore(col_info, invalid_name_leading_underscore):
    col_info['name'] = invalid_name_leading_underscore

    error_re = r"Invalid collection metadata. Expecting 'name' and 'namespace' to not start with '_' but '.*' did"

    log.debug('invalid_name_leading_underscore: %s', invalid_name_leading_underscore)

    with pytest.raises(ValueError,
                       match=error_re) as exc:
        CollectionInfo(**col_info)
    assert invalid_name_leading_underscore in str(exc)


invalid_name_non_alphanumeric = \
    [
     # the '.' in github.com would match the 'no dots' rule first
     'git@github',
     'foo@blip',
     # /, ?, # are non alphanumeric and are invalid
     'rohitggarg/docker-swarm?',
     'c#',
     # leading dash/hyphen and non alpha are invalid
     '--version',
     '-name',
     '-foo',
     # inline '-' is invalid
     'foo-bar',
     'joes-house-of-collections',
     'ansible-galaxy',
     'adfinis-sygroup',
     # double underscores
     'punk__rock',
     'foo__name',
    ]


@pytest.fixture(scope='module',
                params=invalid_name_non_alphanumeric)
def invalid_name_non_alphanumeric(request):
    yield request.param


def test_name_invalid_name_non_alphanumeric(col_info, invalid_name_non_alphanumeric):
    col_info['name'] = invalid_name_non_alphanumeric
    col_info['namespace'] = invalid_name_non_alphanumeric

    log.debug('invalid_name_non_alphanumeric: %s', invalid_name_non_alphanumeric)

    with pytest.raises(ValueError,
                       match=NON_ALPHA_ERROR_PAT) as exc:
        CollectionInfo(**col_info)

    log.debug('exc: %s', str(exc))
    assert invalid_name_non_alphanumeric in str(exc)


def _namespace_invalid_name_non_alphanumeric(col_info, invalid_name_non_alphanumeric):

    log.debug('invalid_name_non_alphanumeric: %s', invalid_name_non_alphanumeric)

    with pytest.raises(ValueError,
                       match=NON_ALPHA_ERROR_PAT) as exc:
        CollectionInfo(**col_info)

    log.debug('exc: %s', str(exc))
    assert invalid_name_non_alphanumeric in str(exc)


invalid_name_leading_number = \
    [
     '0x43',
     '030',
     '2fast2furious',
    ]


@pytest.fixture(scope='module',
                params=invalid_name_leading_number)
def invalid_name_leading_number(request):
    yield request.param


def test_name_invalid_leading_number(col_info, invalid_name_leading_number):
    col_info['name'] = invalid_name_leading_number
    col_info['namespace'] = invalid_name_leading_number

    log.debug('invalid_name_leading_number: %s', invalid_name_leading_number)

    error_re = r"Invalid collection metadata. Expecting 'name' and 'namespace' to not start with a number but '%s' did" % invalid_name_leading_number
    with pytest.raises(ValueError,
                       match=error_re) as exc:
        CollectionInfo(**col_info)
    assert invalid_name_leading_number in str(exc)


invalid_name_uppercase = \
    [
     'EWS',
     'AAROC',
     'AdopteUnOps',
    ]


@pytest.fixture(scope='module',
                params=invalid_name_uppercase)
def invalid_name_uppercase(request):
    yield request.param


def test_name_invalid_uppercase(col_info, invalid_name_uppercase):
    col_info['name'] = invalid_name_uppercase
    col_info['namespace'] = invalid_name_uppercase

    log.debug('invalid_name_uppercase: %s', invalid_name_uppercase)

    with pytest.raises(ValueError,
                       match=NON_ALPHA_ERROR_PAT) as exc:
        CollectionInfo(**col_info)

    log.debug('exc: %s', str(exc))
    assert invalid_name_uppercase in str(exc)


invalid_name_has_dots = \
    [
     'alban.andrieu',
     'dot.net',
     '.',
     '..',
     '_.fsd',
    ]


@pytest.fixture(scope='module',
                params=invalid_name_has_dots)
def invalid_name_has_dots(request):
    yield request.param


def test_name_invalid_has_dots(col_info, invalid_name_has_dots):
    col_info['name'] = invalid_name_has_dots
    col_info['namespace'] = invalid_name_has_dots

    log.debug('invalid_name_has_dots: %s', invalid_name_has_dots)
    error_re = r"Invalid collection metadata. Expecting 'name' and 'namespace' to not include any '\.' but .* has a '\.'"

    with pytest.raises(ValueError,
                       match=error_re) as exc:
        CollectionInfo(**col_info)

    log.debug('exc: %s', str(exc))
    assert invalid_name_has_dots in str(exc)


def test_license_empty_list(col_info):
    col_info['license'] = []

    error_re = r"Valid values for 'license' or 'license_file' are required. But 'license' \(.*\) and 'license_file' \(.*\) were invalid."

    with pytest.raises(ValueError, match=error_re) as exc:
        CollectionInfo(**col_info)

    log.debug('exc: %s', str(exc))


def test_license_valid_and_none_list(col_info):
    col_info['license'] = ['GPL-3.0-or-later', None]

    error_re = r"Invalid collection metadata. Expecting 'license' to be a list of valid SPDX license identifiers, "
    "instead found invalid license identifiers: '.*' in 'license' value .*."

    with pytest.raises(ValueError, match=error_re) as exc:
        CollectionInfo(**col_info)

    log.debug('exc: %s', str(exc))


def test_license_deprecated(col_info):
    col_info['license'] = 'AGPL-1.0'
    res = CollectionInfo(**col_info)
    # Not much to assert, behavior is just a print() side effect
    assert res.license == ['AGPL-1.0']


# TODO maybe... build a text fixture for all of these cases
def test_license_unknown(col_info):
    col_info['license'] = 'SOME-UNKNOWN'
    with pytest.raises(ValueError, match=".*license.*SOME-UNKNOWN.*"):
        CollectionInfo(**col_info)


def test_license_error(col_info):
    col_info['license'] = 'GPLv2'

    with pytest.raises(ValueError) as exc:
        CollectionInfo(**col_info)

    log.debug(str(exc))

    assert 'license' in str(exc)


def test_license_with_valid_license_file(col_info):
    # license=None will be converted to license=[]
    col_info['license'] = None
    col_info['license_file'] = 'MY_LICENSE.txt'

    res = CollectionInfo(**col_info)

    assert res.license_file == 'MY_LICENSE.txt'
    assert res.license == []


def test_license_with_contradicting_license_file(col_info):
    col_info['license_file'] = 'MY_LICENSE.txt'

    res = CollectionInfo(**col_info)

    assert res.license_file == 'MY_LICENSE.txt'
    assert res.license == ['GPL-3.0-or-later']


def test_license_with_none_license_file(col_info):
    col_info['license'] = None
    col_info['license_file'] = None

    error_re = r"Valid values for 'license' or 'license_file' are required. But 'license' \(.*\) and 'license_file' \(.*\) were invalid."
    # error_re = r"Invalid collection metadata. Expecting 'license' to be a list of valid SPDX license identifiers, "
    # "instead found invalid license identifiers: '.*' in 'license' value .*."

    with pytest.raises(ValueError, match=error_re) as exc:
        CollectionInfo(**col_info)

    log.debug('col_info: %s', col_info)
    log.debug(str(exc))


def test_name_required_error(col_info):
    del col_info['name']

    with pytest.raises(ValueError) as exc:
        CollectionInfo(**col_info)
    assert 'name' in str(exc) in str(exc)


def test_namespace_required_error(col_info):
    del col_info['namespace']

    with pytest.raises(ValueError) as exc:
        CollectionInfo(**col_info)
    assert 'namespace' in str(exc) in str(exc)


def test_type_authors_not_list_error(col_info):
    col_info['authors'] = 'chouseknecht'
    with pytest.raises(ValueError) as exc:
        CollectionInfo(**col_info)
    assert 'authors' in str(exc)


def test_tags_non_alpha_error(col_info):
    bad_tag = 'bad-tag!'
    col_info['tags'] = ['goodtag', bad_tag]

    # ValueError: Invalid collection metadata. Expecting tags to contain alphanumeric characters only, instead found 'bad-tag!'.
    error_re = r"Invalid collection metadata. Expecting tags to contain lowercase alphanumeric characters only, instead found '.*'"

    with pytest.raises(ValueError,
                       match=error_re) as exc:
        CollectionInfo(**col_info)
    assert bad_tag in str(exc)


def test_tags_not_a_list_error(col_info):
    not_a_list = 'notataglist'
    col_info['tags'] = not_a_list

    # ValueError: Invalid collection metadata. Expecting 'tags' to be a list
    error_re = r"Invalid collection metadata. Expecting 'tags' to be a list"

    with pytest.raises(ValueError,
                       match=error_re) as exc:
        CollectionInfo(**col_info)
    assert 'tags' in str(exc)


def test_empty():
    col_info = {}
    # ValueError: Invalid collection metadata. 'namespace' is required
    with pytest.raises(ValueError, match=".*'namespace'.*"):
        CollectionInfo(**col_info)


def test_deps_is_none(col_info):
    col_info['dependencies'] = None
    res = CollectionInfo(**col_info)

    log.debug('res: %s', res)

    assert res.dependencies == {}


def test_deps_is_list(col_info):
    col_info['dependencies'] = ['blip', {'sub_dict': 'these_deps_are_wrong'}]
    error_re = r"Invalid collection metadata. Expecting 'dependencies' to be a dict"

    with pytest.raises(ValueError, match=error_re) as exc:
        CollectionInfo(**col_info)

    log.debug(str(exc))


def test_minimal(col_info):
    del col_info['authors']
    res = CollectionInfo(**col_info)

    res.authors.append('Faux Author')
    res.tags.append('sometag')
    res.dependencies.update({'some_dep': '1.1.0',
                             'stuff': '2.2.2'})

    log.debug('res %s res.authors: %s', res, res.authors)

    new_data = col_info.copy()
    res2 = CollectionInfo(**new_data)

    log.debug('res %s res.authors: %s', res, res.authors)
    log.debug('res2 %s res2.authors: %s', res2, res2.authors)

    assert res != res2
    assert res2.authors == []
    assert res2.tags == []
    assert res2.dependencies == {}


def test_authors_append(col_info):
    # let authors go to the defualt
    del col_info['authors']
    res = CollectionInfo(**col_info)
    # log.debug('res %s res.authors: %s', res, res.authors)

    # append to the first objects authors.
    # This should not change the default authors for new
    # CollectionInfo()'s
    res.authors.append('Faux Author')
    log.debug('res %s res.authors: %s', res, res.authors)

    new_data = col_info.copy()

    # No authors provided, should default to []
    res2 = CollectionInfo(**new_data)

    res.authors.append('OnlyAuthoredResNotRes2')

    log.debug('res %s res.authors: %s', res, res.authors)
    log.debug('res2 %s res2.authors: %s', res2, res2.authors)

    # Based on https://www.attrs.org/en/stable/init.html#defaults info about defaults
    # These should not be the same value here
    assert res != res2
    assert res.authors != res2.authors
    assert res2.authors == []
    assert res.authors == ['Faux Author', 'OnlyAuthoredResNotRes2']
    assert 'OnlyAuthoredResNotRes2' not in res2.authors


def test_semantic_version_error(col_info):
    col_info['version'] = 'notaversion'
    with pytest.raises(ValueError) as exc:
        CollectionInfo(**col_info)
    assert 'version' in str(exc)


def test_namespace_property(col_info):
    info = CollectionInfo(**col_info)
    assert info.namespace == 'foo'


def test_label_property(col_info):
    info = CollectionInfo(**col_info)
    assert info.label == 'foo.foo'



import logging
import tempfile

from ansible_galaxy.config import config_file
from ansible_galaxy.config import config
from ansible_galaxy import exceptions

log = logging.getLogger(__name__)

# Using tempfiles here for tests, can revisit and
# mock out file loading, but yaml_file.py does the
# io/stream object stuff so config_file is mostly dealing
# with file load errors.


def test_load_empty():
    yaml_fo = tempfile.NamedTemporaryFile()

    config_data = config_file.load(yaml_fo.name)

    log.debug('config_data: %s', config_data)
    assert config_data is None


def test_load_bogus_path():
    bogus_path = '/dev/null/doesnt/exist.yml'
    config_data = config_file.load(bogus_path)

    log.debug('config_data: %s', config_data)
    assert config_data is None


BAD_YAML = b'''
---
- foo:
    - just a string
    - ['a', 'json', 'list', 'of', 'strings']
# some garbage
 453 56
 d ddddd
:::-
'''


def test_load_busted_yaml():
    yaml_fo = tempfile.NamedTemporaryFile()
    yaml_fo.write(BAD_YAML)
    yaml_fo.flush()

    try:
        config_data = config_file.load(yaml_fo.name)
    except exceptions.GalaxyConfigFileError as e:
        log.debug(e, exc_info=True)
        return

    log.debug('config_data: %s', config_data)
    assert config_data, 'A GalaxyConfigFileError was expected here'


VALID_YAML = b'''
server:
  ignore_certs: true
  url: https://someserver.example.com
content_path: ~/.ansible/some_collections_path/ansible_collections
global_content_path: /usr/local/share/collections/ansible_collections
options:
  role_skeleton_ignore:
    - ^.git$
    - ^.*/.git_keep$
  role_skeleton_path: ~/.some_skeleton
  verbosity: 0
version: 1
'''


def test_load_valid_yaml():
    yaml_fo = tempfile.NamedTemporaryFile()
    yaml_fo.write(VALID_YAML)
    yaml_fo.flush()

    config_data = config_file.load(yaml_fo.name)

    log.debug('config_data: %s', config_data)

    assert config_data is not None
    expected_keys = ['server', 'content_path', 'options']
    for expected_key in expected_keys:
        assert expected_key in config_data

    assert config_data['server']['url'] == 'https://someserver.example.com'
    assert config_data['server']['ignore_certs'] is True
    assert config_data['content_path'] == '~/.ansible/some_collections_path/ansible_collections'
    assert config_data['global_content_path'] == '/usr/local/share/collections/ansible_collections'
    assert config_data['options']['role_skeleton_path'] == '~/.some_skeleton'


def test_save_empty_config():
    yaml_fo = tempfile.NamedTemporaryFile()

    _config = config.Config()

    res = config_file.save(_config.as_dict(), yaml_fo.name)

    log.debug('res: %s', res)


def test_save_config():
    yaml_fo = tempfile.NamedTemporaryFile()

    _config = config.Config()

    config.server = {'url': 'https://someserver.example.com',
                     'ignore_certs': True}
    config.content_path = '~/.ansible/not-content-path'

    res = config_file.save(_config.as_dict(), yaml_fo.name)

    log.debug('res: %s', res)


def test_save_bogus_path():
    _config = config.Config()

    config.server = {'url': 'https://someserver.example.com',
                     'ignore_certs': True}
    config.content_path = '~/.ansible/not-content-path'

    bogus_path = '/dev/null/doesnt/exist.yml'
    res = None
    try:
        res = config_file.save(_config.as_dict(),
                               bogus_path)
    except (OSError, IOError) as e:
        # at the moment, we expect this to raise an OSError or subclass, to
        # verify assert the filename match
        log.debug(e, exc_info=True)
        assert e.filename == bogus_path
        return

    assert res, 'Expected a OSError, IOError or subclass (NotADirectoryError etc) to be raised here'



import json
import logging
import os
import pprint
import tarfile

import attr
import six

from ansible_galaxy.utils import chksums
from ansible_galaxy import collection_members
from ansible_galaxy import collection_artifact_manifest
from ansible_galaxy import collection_artifact_file_manifest
from ansible_galaxy.collection_info import COLLECTION_INFO_FILENAME
from ansible_galaxy.models.collection_artifact_file import CollectionArtifactFile
from ansible_galaxy.models.collection_artifact_manifest import \
    CollectionArtifactManifest
from ansible_galaxy.models.collection_artifact_file_manifest import \
    CollectionArtifactFileManifest
from ansible_galaxy.utils.text import to_bytes

log = logging.getLogger(__name__)

ARCHIVE_FILENAME_TEMPLATE = '{namespace}-{name}-{version}.{extension}'
ARCHIVE_FILENAME_EXTENSION = 'tar.gz'


# TODO: enum
class BuildStatuses(object):
    success = 'SUCCESS'
    incomplete = 'INCOMPLETE'
    failure = 'FAILURE'


@attr.s
class BuildResult(object):
    status = attr.ib()
    manifest = attr.ib(validator=attr.validators.instance_of(CollectionArtifactManifest))
    file_manifest = attr.ib(validator=attr.validators.instance_of(CollectionArtifactFileManifest))
    messages = attr.ib(factory=list)
    errors = attr.ib(factory=list)
    artifact_file_path = attr.ib(default=None)


# # MANIFEST BUILDING STAGE
# find collection members in collection_path/  (roles/*, README*, etc)
#   possibly apply 'pre-discover' include/exclude rules here
# filter members / validate members
#   possibly apply 'post-discover' include/exclude rules here
#
# # ARTIFACT CREATION STAGE
# create a artifact manifest (before creating the archive, since the manifest
#                             will be inside the artifact)
#   sort/order?
#   find file chksums here or in the member finding steps? less stuff here...
#
#
# create a artifact file name artifact_file_name=(v$version.tar.gz)
# persist CollectionArtifact as tar.gz to collection_path/release/$artifact_file_name
#
# # POST BUILD / CLEANUP STAGE
# cleanup anything if needed
# display any useful build results info (name and path to artifact, etc)
#
# Build.run() gathers collection_info, the repo/collection members from disk,
#  then creates a ArtifactManifest(collection_info, repo_members)
#  then creates a CollectionArtifact
#  then persists the collection_artifact
#
# CollectionArtifact will have-a Archive/ArchiveBuilder (generic-ish interface to tarfile for ex)
#
# is a CollectionMember/CollectionMembers object needed? CollectionFileWalker?


def filter_artifact_file_name(attribute, _value):
    '''Used by attr.asdict to remove the src_name attr when serializing'''
    if attribute.name == 'src_name':
        return False
    return True


# TODO: this seems like it should use a strategy pattern...
@attr.s
class Build(object):
    build_context = attr.ib()
    collection_info = attr.ib()

    def run(self, display_callback):

        log.debug('INFO self.collection_info: %s', self.collection_info)

        # ie, 'v1.2.3.tar.gz', not full path
        archive_filename_basename = \
            ARCHIVE_FILENAME_TEMPLATE.format(namespace=self.collection_info.namespace,
                                             name=self.collection_info.name,
                                             version=self.collection_info.version,
                                             extension=ARCHIVE_FILENAME_EXTENSION)

        archive_path = os.path.join(self.build_context.output_path,
                                    archive_filename_basename)
        log.debug('Building archive into archive_path: %s', archive_path)

        # The name of the top level dir in the tar file, ie, there isnt one.
        archive_top_dir = ""
        log.debug('archive_top_dir: %s', archive_top_dir)

        # 'x:gz' is 'create exclusive gzipped'
        tar_file = tarfile.open(archive_path, mode='w:gz')

        # Find collection files, build a file manifest, serialize to json and add to the tarfile
        file_walker = collection_members.FileWalker(collection_path=self.build_context.collection_path)
        col_members = collection_members.CollectionMembers(walker=file_walker)

        log.debug('col_members: %s', col_members)

        col_file_names = col_members.run()
        col_files = collection_artifact_file_manifest.gen_file_manifest_items(col_file_names,
                                                                              self.build_context.collection_path)

        file_manifest = CollectionArtifactFileManifest(files=col_files)

        log.debug('file_manifest: %s', file_manifest)

        for col_member_file in file_manifest.files:
            top_dir = False
            # arcname will be a relative path not an abspath at this point
            rel_path = col_member_file.name or col_member_file.src_name
            if rel_path == '.':
                rel_path = ''
            archive_member_path = rel_path

            log.debug('adding %s to %s (from %s)', archive_member_path,
                      archive_path, col_member_file.name)

            log.debug('name=%s, arcname=%s, top_dir=%s', col_member_file.name, archive_member_path, top_dir)

            # if top_dir:
            #     tar_file.add(col_member_file.name, arcname=archive_top_dir, recursive=False)
            # else:
            #     tar_file.add(col_member_file.name, arcname=archive_member_path, recursive=False)
            tar_file.add(col_member_file.src_name, arcname=archive_member_path, recursive=False)

        # Generate FILES.json contents
        # TODO/FIXME: find and use some streamable file format for the filelist (csv?)
        file_manifest_buf = json.dumps(attr.asdict(file_manifest,
                                                   filter=filter_artifact_file_name),
                                       indent=4)

        log.debug('file_manifest_buf: %s', file_manifest_buf)

        b_file_manifest_buf = to_bytes(file_manifest_buf)
        b_file_manifest_buf_bytesio = six.BytesIO(b_file_manifest_buf)

        archive_manifest_path = collection_artifact_manifest.COLLECTION_MANIFEST_FILENAME
        log.debug('archive_manifest_path: %s', archive_manifest_path)

        archive_file_manifest_path = collection_artifact_file_manifest.COLLECTION_FILE_MANIFEST_FILENAME
        log.debug('archive_file_manifest_path: %s', archive_file_manifest_path)

        file_manifest_tar_info = tar_file.gettarinfo(os.path.join(self.build_context.collection_path, COLLECTION_INFO_FILENAME))

        file_manifest_tar_info.name = archive_file_manifest_path
        file_manifest_tar_info.size = len(b_file_manifest_buf)

        # Add FILES.json contents to tarball
        tar_file.addfile(tarinfo=file_manifest_tar_info,
                         fileobj=b_file_manifest_buf_bytesio)

        # addfile reads to end of bytesio, seek back to begin
        b_file_manifest_buf_bytesio.seek(0)
        file_manifest_file_chksum = chksums.sha256sum_from_fo(b_file_manifest_buf_bytesio)

        log.debug('file_manifest_file_chksum: %s', file_manifest_file_chksum)

        # file_manifest_file_name_in_archive = os.path.relpath(archive_file_manifest_path, self.build_context.collection_path)

        file_manifest_file_item = CollectionArtifactFile(src_name=collection_artifact_file_manifest.COLLECTION_FILE_MANIFEST_FILENAME,
                                                         # The path where the file will live inside the archive
                                                         name=collection_artifact_file_manifest.COLLECTION_FILE_MANIFEST_FILENAME,
                                                         ftype='file',
                                                         chksum_type='sha256',
                                                         chksum_sha256=file_manifest_file_chksum)

        # Generage MANIFEST.json contents
        manifest = CollectionArtifactManifest(collection_info=self.collection_info,
                                              file_manifest_file=file_manifest_file_item)

        log.debug('manifest: %s', manifest)

        manifest_buf = json.dumps(attr.asdict(manifest,
                                              filter=filter_artifact_file_name),
                                  # sort_keys=True,
                                  indent=4)
        log.debug('manifest buf: %s', manifest_buf)

        # add MANIFEST.yml to archive
        b_manifest_buf = to_bytes(manifest_buf)
        b_manifest_buf_bytesio = six.BytesIO(b_manifest_buf)

        archive_manifest_path = os.path.join(archive_top_dir,
                                             collection_artifact_manifest.COLLECTION_MANIFEST_FILENAME)
        log.debug('archive_manifest_path: %s', archive_manifest_path)

        # copy the uid/gid/perms for galaxy.yml to use on the manifes. Need sep instances for manifest and file_manifest
        # TODO: decide on what the generators owner/group/perms should be (root.root 644?)
        manifest_tar_info = tar_file.gettarinfo(os.path.join(self.build_context.collection_path, COLLECTION_INFO_FILENAME))

        manifest_tar_info.name = archive_manifest_path
        manifest_tar_info.size = len(b_manifest_buf)

        # TODO: set mtime equal to the 'build time' / build_info when we start creating that.

        tar_file.addfile(tarinfo=manifest_tar_info,
                         fileobj=b_manifest_buf_bytesio)

        log.debug('populated tarfile %s: %s', archive_path,
                  pprint.pformat(tar_file.getmembers))

        tar_file.close()

        # could in theory make creating the release artifact work much the same
        # as serializing some object (I mean, that is what it is... but

        messages = ['Building collection: %s' % self.build_context.collection_path,
                    'Created  artifact: %s' % archive_path]

        result = BuildResult(status=BuildStatuses.success,
                             messages=messages,
                             # errors=[],
                             errors=col_members.walker.file_errors,
                             manifest=manifest,
                             file_manifest=file_manifest,
                             artifact_file_path=archive_path)

        for message in result.messages:
            log.info(message)
            display_callback(message)

        for error in result.errors:
            log.error(error)
            display_callback(error, level='warning')

        return result




import hashlib
import json
import logging
import os
import tarfile

from ansible_galaxy.rest_api import GalaxyAPI
from ansible_galaxy import collection_artifact_manifest
from ansible_galaxy import exceptions
from ansible_galaxy.utils.text import to_text

log = logging.getLogger(__name__)


def _get_file_checksum(file_path):
    checksum = hashlib.sha256()
    with open(file_path, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            checksum.update(byte_block)
    return checksum.hexdigest()


def _publish(galaxy_context,
             archive_path,
             publish_api_key=None,
             display_callback=None):

    results = {
        'errors': [],
        'success': True
    }

    archive = tarfile.open(archive_path, 'r')
    top_dir = os.path.commonprefix(archive.getnames())
    manifest_path = os.path.join(top_dir,
                                 collection_artifact_manifest.COLLECTION_MANIFEST_FILENAME)
    try:
        manifest_file = archive.extractfile(manifest_path)
    except tarfile.TarError as exc:
        raise exceptions.GalaxyPublishError(str(exc), archive_path=archive_path)

    try:
        manifest = collection_artifact_manifest.load(manifest_file)
    except Exception as exc:
        raise exceptions.GalaxyPublishError(str(exc), archive_path=archive_path)

    # display_callback('Creating publish task for %s from artifact archive %s' %
    #                 (manifest.collection_info.label, archive_path))
    # display_callback(json.dumps(attr.asdict(manifest.collection_info)))

    api = GalaxyAPI(galaxy_context)

    collection_name = manifest.collection_info.name

    data = {
        'sha256': _get_file_checksum(archive_path),
        'name': collection_name,
        'version': manifest.collection_info.version
    }

    log.debug("Publishing file %s with data: %s" % (archive_path, json.dumps(data)))

    b_response_body = api.publish_file(data, archive_path, publish_api_key)
    response_body = to_text(b_response_body, errors='surrogate_or_strict')

    # TODO: try/except on json error but let bubble up for now

    response_data = json.loads(response_body)
    results['response_data'] = response_data

    return results


def publish(galaxy_context, archive_path, publish_api_key, display_callback):

    results = _publish(galaxy_context,
                       archive_path,
                       publish_api_key=publish_api_key,
                       display_callback=display_callback)

    log.debug('cli publish action results: %s', json.dumps(results))

    if results['errors']:
        for error in results['errors']:
            display_callback(error)

    if results['success']:
        if results['response_data'].get('task', None):
            display_callback('Publish task for %s created at %s/%s' %
                             (archive_path,
                              galaxy_context.server['url'],
                              results['response_data']['task']))
        return os.EX_OK  # 0

    return os.EX_SOFTWARE  # 70



import logging
import os

import semantic_version

from ansible_galaxy import repository
from ansible_galaxy import repository_archive
from ansible_galaxy.models.repository import Repository
from ansible_galaxy.models.repository_spec import RepositorySpec
from ansible_galaxy.actions import build
from ansible_galaxy.models.build_context import BuildContext

log = logging.getLogger(__name__)


def display_callback(msg, **kwargs):
    log.debug(msg)


def build_repo_artifact(galaxy_context_, tmp_dir):
    output_path = tmp_dir.mkdir('mazer_test_build_action_test_build')

    collection_path = os.path.join(os.path.dirname(__file__), 'collection_examples/hello')
    build_context = BuildContext(collection_path, output_path=output_path.strpath)
    ret = build._build(galaxy_context_, build_context, display_callback)

    log.debug('ret: %s', ret)

    return ret


def test_load_from_archive_artifact(galaxy_context, tmpdir):
    built_res = build_repo_artifact(galaxy_context, tmpdir)
    archive_path = built_res['build_results'].artifact_file_path

    repo_archive = repository_archive.load_archive(archive_path)
    res = repository.load_from_archive(repo_archive)

    assert isinstance(res, Repository)
    assert isinstance(res.repository_spec, RepositorySpec)
    assert isinstance(res.repository_spec.version, semantic_version.Version)

    assert res.repository_spec.namespace == 'greetings_namespace'
    assert res.repository_spec.name == 'hello'


# def load_from_dir(content_dir, namespace, name, installed=True):
def test_load_from_dir_no_dir():
    res = repository.load_from_dir('/dev/null/doesntexist', 'some_namespace', 'some_name')

    log.debug('res: %s', res)

    assert res is None, 'load_from_dir() on a dir that does not exist should have returned None'


def test_load_dict():
    data = '''{"foo": "bar"}'''

    res = repository.load(data)

    log.debug('res: %s', res)

    assert isinstance(res, dict)
    assert res['foo'] == 'bar'



import logging

import pytest

from ansible_galaxy.models.collection_artifact_file_manifest import CollectionArtifactFileManifest

log = logging.getLogger(__name__)


def test_example():
    files = [{'name': 'roles/some_role/defaults/main.yml',
              'chksum_type': 'sha256',
              'chksum_sha256': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',
              'ftype': 'file'}]

    file_manifest = CollectionArtifactFileManifest(files=files)

    log.debug('file_manifest: %s', file_manifest)

    assert isinstance(file_manifest, CollectionArtifactFileManifest)
    assert len(file_manifest.files) == 1


def test_init_no_files():
    files = []
    file_manifest = CollectionArtifactFileManifest(files=files)

    log.debug('file_manifest: %s', file_manifest)

    assert isinstance(file_manifest, CollectionArtifactFileManifest)
    assert len(file_manifest.files) == 0


def test_init_one_empty_file_item():
    files = [{}]

    # Note; KeyError from the converter, not a ValueError from init
    with pytest.raises(KeyError):
        CollectionArtifactFileManifest(files=files)



import logging

from ansible_galaxy import exceptions
from ansible_galaxy.fetch import galaxy_url
from ansible_galaxy.fetch import local_file
from ansible_galaxy.fetch import remote_url
from ansible_galaxy.fetch import scm_url
from ansible_galaxy.fetch import editable
from ansible_galaxy.models.repository_spec import FetchMethods

log = logging.getLogger(__name__)


def get(galaxy_context, requirement_spec):
    """determine how to download a repo, builds a fetch instance, and returns the instance"""

    fetcher = None

    # FIXME: note that ignore_certs for the galaxy
    # server(galaxy_context.server['ignore_certs'])
    # does not really imply that the repo archive download should ignore certs as well
    # (galaxy api server vs cdn) but for now, we use the value for both

    if requirement_spec.fetch_method == FetchMethods.EDITABLE:
        fetcher = editable.EditableFetch(requirement_spec=requirement_spec,
                                         galaxy_context=galaxy_context)
    elif requirement_spec.fetch_method == FetchMethods.SCM_URL:
        fetcher = scm_url.ScmUrlFetch(requirement_spec=requirement_spec)
    elif requirement_spec.fetch_method == FetchMethods.LOCAL_FILE:
        # the file is a tar, so open it that way and extract it
        # to the specified (or default) content directory
        fetcher = local_file.LocalFileFetch(requirement_spec)
    elif requirement_spec.fetch_method == FetchMethods.REMOTE_URL:
        fetcher = remote_url.RemoteUrlFetch(requirement_spec=requirement_spec,
                                            validate_certs=not galaxy_context.server['ignore_certs'])
    elif requirement_spec.fetch_method == FetchMethods.GALAXY_URL:
        fetcher = galaxy_url.GalaxyUrlFetch(requirement_spec=requirement_spec,
                                            galaxy_context=galaxy_context)
    else:
        raise exceptions.GalaxyError('No approriate content fetcher found for %s %s',
                                     requirement_spec.scm, requirement_spec.src)

    log.debug('Using fetcher: %s for requirement_spec: %r', fetcher, requirement_spec)

    return fetcher




import logging

from ansible_galaxy.utils.repository_name import repo_url_to_repo_name

log = logging.getLogger(__name__)


# FIXME: more of a view, but baby steps
# FIXME: likely needs to learn version=1,name='blip', etc. And tests
def role_spec_parse(role_spec):
    # takes a repo and a version like
    # git+http://git.example.com/repos/repo.git,v1.0
    # and returns a list of properties such as:
    # {
    #   'scm': 'git',
    #   'src': 'http://git.example.com/repos/repo.git',
    #   'version': 'v1.0',
    #   'name': 'repo'
    # }
    log.warning("The comma separated role spec format, use the yaml/explicit format instead. Line that trigger this: %s (version='2.7')", role_spec)

    default_role_versions = dict(git='master', hg='tip')

    role_spec = role_spec.strip()
    role_version = ''
    if role_spec == "" or role_spec.startswith("#"):
        return (None, None, None, None)

    tokens = [s.strip() for s in role_spec.split(',')]

    # assume https://github.com URLs are git+https:// URLs and not
    # tarballs unless they end in '.zip'
    if 'github.com/' in tokens[0] and not tokens[0].startswith("git+") and not tokens[0].endswith('.tar.gz'):
        tokens[0] = 'git+' + tokens[0]

    if '+' in tokens[0]:
        (scm, role_url) = tokens[0].split('+')
    else:
        scm = None
        role_url = tokens[0]

    if len(tokens) >= 2:
        role_version = tokens[1]

    if len(tokens) == 3:
        role_name = tokens[2]
    else:
        role_name = repo_url_to_repo_name(tokens[0])

    if scm and not role_version:
        role_version = default_role_versions.get(scm, '')

    return dict(scm=scm, src=role_url, version=role_version, name=role_name)



import glob
import itertools
import logging
import os

from ansible_galaxy import installed_repository_db
from ansible_galaxy import matchers
from ansible_galaxy.models.content_item import ContentItem

log = logging.getLogger(__name__)


def python_content_path_iterator(repository, content_item_sub_dir):
    '''For paths where python plugins live, filters out __init__.py'''

    return (x for x in glob.iglob('%s/%s/*' % (repository.path, content_item_sub_dir)) if not x.endswith('__init__.py'))


def glob_content_path_iterator(repository, content_item_sub_dir):
    '''For paths content_item_sub_dir/*, like modules'''
    return glob.iglob('%s/%s/*' % (repository.path, content_item_sub_dir))


def is_plugin_dir(plugin_dir_path):
    '''Check plugins/ items to see if a potential plugin dir is a dir

    No other checks are used to determine if plugin_dir_path is actually a plugin dir'''
    log.debug('plugin_dir_path: %s', plugin_dir_path)

    if os.path.isdir(plugin_dir_path):
        return True

    return False


def plugin_content_item_types(repository):
    '''Return all the subdirs in plugins/ assuming only plugin type dirs exist

    ie, this doesn't have preconceived notions of the possible plugin types'''
    plugins_dir = os.path.join(repository.path, 'plugins')
    try:
        res = [x for x in os.listdir(plugins_dir) if is_plugin_dir(os.path.join(plugins_dir, x))]
        # log.debug('plugin_content_item_types: %s', res)
        return res
    except (OSError, IOError):
        # log.warning(e)
        return []


# need a content_type matcher?
def repository_content_iterator(repository, content_item_type, content_item_paths_iterator):

    for content_item_path in content_item_paths_iterator:
        repo_namespace = repository.repository_spec.namespace
        file_name = os.path.basename(content_item_path)

        # This assumes module and plugins with filenames with an extension will always
        # use the content item name as the pre-ext part of filename. That could change
        # in the future.
        content_item_name = os.path.splitext(file_name)[0]

        content_item = ContentItem(namespace=repo_namespace,
                                   name=content_item_name,
                                   path=file_name,
                                   content_item_type=content_item_type,
                                   version=repository.repository_spec.version)

        log.debug('Found %s "%s" at %s', content_item, content_item.name, file_name)
        yield content_item


def all_content_item_types_iterator(repository):
    all_content_iterators = [
        repository_content_iterator(repository,
                                    'roles',
                                    glob_content_path_iterator(repository, 'roles')),
        repository_content_iterator(repository,
                                    'modules',
                                    glob_content_path_iterator(repository, 'modules')),
        repository_content_iterator(repository,
                                    'module_utils',
                                    python_content_path_iterator(repository, 'module_utils')),
    ]

    for plugin_content_item_type in plugin_content_item_types(repository):
        plugin_iterator = repository_content_iterator(repository,
                                                      plugin_content_item_type,
                                                      python_content_path_iterator(repository,
                                                                                   os.path.join('plugins', plugin_content_item_type)))
        all_content_iterators.append(plugin_iterator)

    # chain all the iterables that generator ContentItems together
    return itertools.chain.from_iterable(all_content_iterators)


def installed_content_item_iterator(galaxy_context,
                                    namespace_match_filter=None,
                                    repository_spec_match_filter=None,
                                    content_item_match_filter=None,
                                    content_item_type=None):

    # match_all works for all types
    namespace_match_filter = namespace_match_filter or matchers.MatchAll()
    repository_spec_match_filter = repository_spec_match_filter or matchers.MatchAll()
    content_item_match_filter = content_item_match_filter or matchers.MatchAll()

    content_item_type = content_item_type or 'roles'

    installed_repo_db = installed_repository_db.InstalledRepositoryDatabase(galaxy_context)

    # for namespace_full_path in namespace_paths_iterator:
    for installed_repository in installed_repo_db.select(namespace_match_filter=namespace_match_filter,
                                                         repository_spec_match_filter=repository_spec_match_filter):
        log.debug('Found repository "%s" at %s', installed_repository.repository_spec.label, installed_repository.path)
        installed_repository_full_path = installed_repository.path

        if not repository_spec_match_filter(installed_repository):
            log.debug('The repository_match_filter %s failed to match for %s', repository_spec_match_filter, installed_repository)
            continue

        all_content_iterator = all_content_item_types_iterator(installed_repository)

        log.debug('Looking for %s in repository at %s', content_item_type, installed_repository_full_path)

        for installed_content_item in all_content_iterator:
            log.debug('installed_content_item: %s', installed_content_item)

            if not content_item_match_filter(installed_content_item):
                log.debug('%s was not matched by content_match_filter: %s', installed_content_item, content_item_match_filter)
                continue

            # this is sort of the 'join' of installed_repository and installed_content
            content_info = {'path': installed_content_item.path,
                            'content_data': installed_content_item,
                            'installed_repository': installed_repository,
                            'version': installed_content_item.version,
                            }

            yield content_info


class InstalledContentItemDatabase(object):

    def __init__(self, installed_context=None):
        self.installed_context = installed_context

    # select content based on matching the installed namespace.repository and/or the content itself
    def select(self, repository_spec_match_filter=None, content_match_filter=None):
        # ie, default to select * more or less
        repository_spec_match_filter = repository_spec_match_filter or matchers.MatchAll()
        content_match_filter = content_match_filter or matchers.MatchAll()

        roles_content_item_iterator = \
            installed_content_item_iterator(self.installed_context,
                                            content_item_type='roles',
                                            repository_spec_match_filter=repository_spec_match_filter,
                                            content_item_match_filter=content_match_filter)

        for matched_content_item in roles_content_item_iterator:
            yield matched_content_item



from collections import OrderedDict
import logging

from ansible_galaxy import yaml_persist
from ansible_galaxy.models.collection_info import CollectionInfo

log = logging.getLogger(__name__)


test_dict = {'blip': 1,
             'bar': [{'a': 'A', 'b': 'B', 'cc': [None, 1, 2.3, ('blip', 'foo')]},
                     {'other': 'stuff'}],
             'somelists': [
                 [11, 12, 13, 14],
                 [555, 1.1, ['x', 't', 'z']],
             ],
             'somedicts': {'some_o_dict': OrderedDict({'3a': '3A', '3b': '3B'}),
                           'level2': {'ccvb': 'bcvvvv'},
                           'level3': {'hfgfg': '4e56'},
                           },
             'a_tuple': ('j', 'k', 'l', 'm'),
             }


def test_safe_dump():
    yaml_str = yaml_persist.safe_dump(test_dict)
    log.debug('yaml_str:\n%s', yaml_str)

    assert yaml_str is not None
    assert 'somedicts' in yaml_str


def test_safe_dump_attr():
    col_info = CollectionInfo(namespace='some_ns',
                              name='some_name',
                              version='1.2.3',
                              license='MIT',
                              description='stuff',
                              authors=['Charles Oakley'],
                              tags=['sometag', 'anothertag'],
                              dependencies=OrderedDict([('foo.bar', '*'),
                                                        ('blip.bar', '>=1.1.0')]))

    log.debug('col_info: %s', col_info)
    yaml_str = yaml_persist.safe_dump(col_info)
    log.debug('yaml_str:\n%s', yaml_str)

    assert yaml_str is not None
    assert 'blip.bar' in yaml_str



import logging
import os

from ansible_galaxy import exceptions
from ansible_galaxy.actions import list as list_action

log = logging.getLogger(__name__)


def display_callback(msg, **kwargs):
    log.debug(msg)


# TODO: make a fixture that sets up a faux installed_*_db

def test__list(galaxy_context, mocker):
    # galaxy_context.content_path = os.path.join(galaxy_context.content_path, 'doesntexist')
    mocker.patch('ansible_galaxy.installed_namespaces_db.get_namespace_paths',
                 return_value=iter(['ns_blip', 'ns_foo']))
    mocker.patch('ansible_galaxy.installed_repository_db.get_repository_paths',
                 return_value=iter(['n_bar', 'n_baz']))

    mocker.patch('ansible_galaxy.repository.os.path.isdir',
                 return_value=True)

    mocker.patch('ansible_galaxy.installed_content_item_db.glob_content_path_iterator',
                 return_value=iter(['/dev/null/content/ns_foo/n_bar/roles/role-1',
                                    '/dev/null/content/ns_blip/n_bar/roles/role-3',
                                    '/dev/null/content/ns_blip/n_baz/roles/role-2']))

    res = list_action._list(galaxy_context,
                            display_callback=display_callback)

    # log.debug('res: %s', pprint.pformat(res))

    assert isinstance(res, list)
    assert 'ns_blip.n_bar' in [x['installed_repository'].repository_spec.label for x in res]


def test_list_empty_roles_paths(galaxy_context):

    try:
        list_action.list_action(galaxy_context,
                                display_callback=display_callback)
    except exceptions.GalaxyError as e:
        log.debug(e, exc_info=True)
        raise


def test_list_no_content_dir(galaxy_context):
    galaxy_context.content_path = os.path.join(galaxy_context.content_path, 'doesntexist')
    res = list_action.list_action(galaxy_context,
                                  display_callback=display_callback)

    # TODO: list should probably return non-zero if galaxy_context.content_path doesnt exist,
    #       but should probaly initially check that when creating galaxy_context
    assert res == 0



########################################################################
#
# (C) 2018, Ansible by Red Hat
#
# This file is part of Mazer
#
# Mazer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Mazer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible. If not, see <http://www.gnu.org/licenses/>.
#
########################################################################

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import json
import logging
import os
import sys

from ansible_galaxy.actions import build
from ansible_galaxy.actions import info
from ansible_galaxy.actions import install
from ansible_galaxy.actions import list as list_action
from ansible_galaxy.actions import migrate_role
from ansible_galaxy.actions import remove
from ansible_galaxy.actions import version
from ansible_galaxy.actions import publish
from ansible_galaxy.config import defaults
from ansible_galaxy.config import config

from ansible_galaxy import matchers
from ansible_galaxy import rest_api

from ansible_galaxy.models.context import GalaxyContext
from ansible_galaxy.models.build_context import BuildContext
from ansible_galaxy.models.migrate_role_context import MigrateRoleContext

from ansible_galaxy_cli import cli
from ansible_galaxy_cli import __version__ as galaxy_cli_version
from ansible_galaxy_cli import exceptions as cli_exceptions

log = logging.getLogger(__name__)


def exit_without_ignore(ignore_errors, msg=None, rc=1):
    """
    Exits with the specified return code unless the
    option --ignore-errors was specified
    """
    ignore_error_blurb = '- you can use --ignore-errors to skip failed content items and finish processing the list.'
    if not ignore_errors:
        message = ignore_error_blurb
        if msg:
            message = '%s:\n%s' % (msg, ignore_error_blurb)
        raise cli_exceptions.GalaxyCliError(message)


def get_config_path_from_env():
    for env_var in ('MAZER_CONFIG', 'ANSIBLE_GALAXY_CONFIG'):
        raw_config_file_path = os.environ.get(env_var, None)

        if raw_config_file_path:
            log.info("Using config file '%s' as specified by env var '%s'",
                     raw_config_file_path, env_var)
            return raw_config_file_path

    return None


class GalaxyCLI(cli.CLI):
    SKIP_INFO_KEYS = ("name", "description", "readme_html", "related", "summary_fields", "average_aw_composite", "average_aw_score", "url")
    VALID_ACTIONS = ("build", "info", "install", "list", "migrate_role", "publish", "remove", "version")
    VALID_ACTION_ALIASES = {'content-install': 'install'}

    def __init__(self, args):
        self.api = None
        super(GalaxyCLI, self).__init__(args)

    def set_action(self):

        super(GalaxyCLI, self).set_action()

        if self.action == "build":
            self.parser.set_usage("usage: %prog build [options]")
            self.parser.add_option('--collection-path', dest='collection_path', default=None,
                                   help='The path in which the collection is located. The default is the current working directory.')
            self.parser.add_option('--output-path', dest='output_path', default=None,
                                   help='The path in which the collection artifact will be created. The default is ./releases/.')
        if self.action == "publish":
            self.parser.set_usage("usage: %prog publish [options] archive_path")
            # TODO: Include the url for the pref page where the API key can be found, however, that
            #       needs to know the server url which isn't known until after cli args are parsed.
            self.parser.add_option('--api-key', dest='publish_api_key', action='store', default=None,
                                   help='The Galaxy API key.')
        if self.action == "info":
            self.parser.set_usage("usage: %prog info [options] repo_name[,version]")
        elif self.action == "install":
            self.parser.set_usage("usage: %prog install [options] [-r FILE | repo_name(s)[,version] | scm+repo_url[,version] | tar_file(s)]")
            self.parser.add_option('-g', '--global', dest='global_install', action='store_true',
                                   help='Install content to the path containing your global or system-wide content. The default is the '
                                   'global_content_path configured in your mazer.yml file (/usr/share/ansible/content, if not configured)')
            self.parser.add_option('-e', '--editable', dest='editable_install', action='store_true',
                                   help='Link a local directory into the content path for development and testing')
            self.parser.add_option('-i', '--ignore-errors', dest='ignore_errors', action='store_true', default=False,
                                   help='Ignore errors and continue with the next specified repo.')
            self.parser.add_option('-n', '--no-deps', dest='no_deps', action='store_true', default=False, help='Don\'t download roles listed as dependencies')
            self.parser.add_option('-r', '--role-file', dest='role_file', help='A file containing a list of roles to be imported')
            self.parser.add_option('--namespace', dest='namespace', default=None,
                                   help='The namespace to use when installing content (required for installs from local scm repo or archives)')
        elif self.action == "remove":
            self.parser.set_usage("usage: %prog remove repo1 repo2 ...")
        elif self.action == "list":
            self.parser.set_usage("usage: %prog list [repo_name]")
            self.parser.add_option('--content', dest='list_content', default=False, action='store_true', help="List each content item type in a repo")
        elif self.action == "version":
            self.parser.set_usage("usage: %prog version")

        # options that apply to more than one action
        if self.action in ("info",):
            self.parser.add_option('--offline', dest='offline', default=False, action='store_true', help="Prevent Mazer from calling the galaxy API.")

        if self.action not in ("publish", "version",):
            # NOTE: while the option type=str, the default is a list, and the
            # callback will set the value to a list.
            self.parser.add_option('-C', '--content-path', dest='content_path',
                                   help='The path to the directory containing your Galaxy content. The default is the content_path configured in your'
                                        'mazer.yml file (~/.ansible/content, if not configured)', type='str')

        if self.action == "migrate_role":
            self.parser.add_option('--role', dest='role_convertee_path',
                                   help='The path to the role that will be converted to a collection')
            self.parser.add_option('--output-dir', dest='collection_output_dir',
                                   help='The path to write the new collection to')
            self.parser.add_option('--force', dest='output_force', action='store_true', default=False,
                                   help='Write to the output dir even if parts of it already exists')
            self.parser.add_option('--namespace', dest='collection_namespace',
                                   help='The namespace to use for the new collection')
            self.parser.add_option('--name', dest='collection_name',
                                   help='The name to use for the new collection')
            self.parser.add_option('--version', dest='collection_version',
                                   help='The version to use for the new collection')
            self.parser.add_option('--license', dest='collection_license', default=None,
                                   help='The SPDX license identifier to use for the new collection. For ex, "GPL-3.0-or-later", "MIT", "BSD-3-Clause"')

        if self.action in ("install",):
            self.parser.add_option('-f', '--force', dest='force', action='store_true', default=False, help='Force overwriting an existing collection')

    def parse(self):
        ''' create an options parser for bin/ansible '''

        self.parser = cli.CLI.base_parser(
            usage="usage: %%prog [%s] [--help] [options] ..." % "|".join(self.VALID_ACTIONS),
            epilog="\nSee '%s <command> --help' for more information on a specific command.\n\n" % os.path.basename(sys.argv[0])
        )

        # common
        self.parser.add_option('-s', '--server', dest='server_url', default=None, help='The API server destination')

        self.parser.add_option('-c', '--ignore-certs', action='store_true', dest='ignore_certs', default=None,
                               help='Ignore SSL certificate validation errors.')
        self.set_action()

        super(GalaxyCLI, self).parse()

        if self.action == 'install' and getattr(self.options, 'content_path') and getattr(self.options, 'global_install'):
            raise cli_exceptions.CliOptionsError('--content-path and --global are mutually exclusive')

    def _get_galaxy_context(self, options, config):
        # use content_path from options if availble but fallback to configured content_path
        options_content_path = None
        if hasattr(options, 'content_path'):
            options_content_path = options.content_path

        raw_content_path = options_content_path or config.content_path

        if hasattr(options, 'global_install') and options.global_install:
            raw_content_path = config.global_content_path

        content_path = os.path.expanduser(raw_content_path)

        server = config.server.copy()

        if getattr(options, 'server_url', None):
            server['url'] = options.server_url

        if getattr(options, 'ignore_certs', None):
            # use ignore certs from options if available, but fallback to configured ignore_certs
            server['ignore_certs'] = options.ignore_certs

        galaxy_context = GalaxyContext(server=server, content_path=content_path)

        return galaxy_context

    def run(self):

        raw_config_file_path = get_config_path_from_env() or defaults.CONFIG_FILE

        self.config_file_path = os.path.abspath(os.path.expanduser(raw_config_file_path))

        super(GalaxyCLI, self).run()

        self.config = config.load(self.config_file_path)

        log.debug('configuration: %s', json.dumps(self.config.as_dict(), indent=None))

        # cli --server value or the url field of the first server in config
        # TODO: pass list of server config objects to GalaxyContext and/or create a GalaxyContext later
        # server_url = self.options.server_url or self.config['servers'][0]['url']
        # ignore_certs = self.options.ignore_certs or self.config['servers'][0]['ignore_certs']

        galaxy_context = self._get_galaxy_context(self.options, self.config)

        log.debug('galaxy context: %s', galaxy_context)

        self.api = rest_api.GalaxyAPI(galaxy_context)

        log.debug('execute action: %s', self.action)
        log.debug('execute action with options: %s', self.options)
        log.debug('execute action with args: %s', self.args)

        return self.execute()

    def execute_build(self):
        """
        Create a collection artifact from a collection src repository.
        """

        log.debug('options: %s', self.options)
        log.debug('args: %s', self.args)

        galaxy_context = self._get_galaxy_context(self.options, self.config)

        # default to looking for collection in cwd if not specified
        cwd = os.getcwd()
        collection_path = self.options.collection_path or cwd
        # write the output archive to output_path, which will default to collection
        # relative releases/
        output_path = self.options.output_path or os.path.join(collection_path, 'releases')

        build_context = BuildContext(collection_path=collection_path,
                                     output_path=output_path)

        return build.build(galaxy_context,
                           build_context,
                           display_callback=self.display)

    def execute_info(self):
        """
        Display detailed information about an installed collection, as well as info available from the Galaxy API.
        """

        if len(self.args) == 0:
            # the user needs to specify a role
            raise cli_exceptions.CliOptionsError("- you must specify a user/role name")

        log.debug('args=%s', self.args)

        galaxy_context = self._get_galaxy_context(self.options, self.config)

        repository_spec_strings = self.args

        # FIXME: rc?
        return info.info_repository_specs(galaxy_context, self.api, repository_spec_strings,
                                          display_callback=self.display,
                                          offline=self.options.offline)

    def execute_install(self):
        """
        Install a collection.
        """

        self.log.debug('self.options: %s', self.options)

        galaxy_context = self._get_galaxy_context(self.options, self.config)
        requested_spec_strings = self.args

        # TODO: build requirement_specs from requested_collection_specs strings
        rc = install.install_repository_specs_loop(galaxy_context,
                                                   editable=self.options.editable_install,
                                                   repository_spec_strings=requested_spec_strings,
                                                   namespace_override=self.options.namespace,
                                                   display_callback=self.display,
                                                   ignore_errors=self.options.ignore_errors,
                                                   no_deps=self.options.no_deps,
                                                   force_overwrite=self.options.force)

        return rc

    def execute_publish(self):
        """
        Publish a collection artifact to Galaxy.
        """

        galaxy_context = self._get_galaxy_context(self.options, self.config)

        if not len(self.args) or not os.path.isfile(self.args[0]):
            raise cli_exceptions.CliOptionsError("- you must specify a path to a collection archive")

        return publish.publish(galaxy_context,
                               self.args[0],
                               self.options.publish_api_key,
                               display_callback=self.display)

    def execute_remove(self):
        """
        Remove a list of repository from the local system.
        """

        if len(self.args) == 0:
            raise cli_exceptions.CliOptionsError('- you must specify at least one role to remove.')

        galaxy_context = self._get_galaxy_context(self.options, self.config)

        if self.args:
            match_filter = matchers.MatchLabels(self.args)

        return remove.remove(galaxy_context,
                             repository_spec_match_filter=match_filter,
                             display_callback=self.display)

    def execute_list(self):
        """
        List repositories, roles, and collections installed on the local file system.
        """

        galaxy_context = self._get_galaxy_context(self.options, self.config)

        match_filter = matchers.MatchAll()

        list_content = self.options.list_content
        if self.args:
            match_filter = matchers.MatchNamespacesOrLabels(self.args)

        return list_action.list_action(galaxy_context,
                                       repository_spec_match_filter=match_filter,
                                       list_content=list_content,
                                       display_callback=self.display)

    def execute_version(self):
        return version.version(config_file_path=self.config_file_path,
                               cli_version=galaxy_cli_version,
                               display_callback=self.display)

    def execute_migrate_role(self):

        migrate_role_context = MigrateRoleContext(role_path=self.options.role_convertee_path,
                                                  output_path=self.options.collection_output_dir,
                                                  collection_namespace=self.options.collection_namespace,
                                                  collection_name=self.options.collection_name,
                                                  collection_version=self.options.collection_version,
                                                  collection_license=self.options.collection_license,
                                                  output_force=self.options.output_force)

        return migrate_role.migrate(migrate_role_context=migrate_role_context,
                                    display_callback=self.display)




import logging
import os

from ansible_galaxy.fetch import editable
from ansible_galaxy import requirements
from ansible_galaxy.models.requirement_spec import RequirementSpec

log = logging.getLogger(__name__)


def test_editable_fetch_find(galaxy_context, mocker, tmpdir):
    name = 'mazer_fetch_test_editable'
    namespace_override = 'some_editable_namespace'

    tmp_path = tmpdir.mkdir(name)

    more_reqs = requirements.from_requirement_spec_strings([tmp_path.strpath],
                                                           namespace_override=namespace_override,
                                                           editable=True)

    import pprint
    log.debug('more_reqs: %s', pprint.pformat(more_reqs))

    req_spec = more_reqs[0].requirement_spec

    fetcher = editable.EditableFetch(galaxy_context, req_spec)

    log.debug(fetcher)

    res = fetcher.find()
    log.debug('res: %s', res)

    assert isinstance(res, dict)
    assert res['content']['galaxy_namespace'] == namespace_override
    assert res['content']['repo_name'] == name
    assert res['custom']['real_path'] == tmp_path.strpath


def test_editable_fetch_fetch(galaxy_context, mocker, tmpdir):
    name = 'mazer_fetch_test_editable'
    namespace_override = 'some_editable_namespace'

    tmp_working_path = tmpdir.mkdir('some_working_tree')
    dest_tmp_path = tmp_working_path.mkdir('some_checkout')

    req_spec = RequirementSpec(namespace=namespace_override,
                               name=name,
                               fetch_method='EDITABLE',
                               src=dest_tmp_path,
                               version_spec='*')
    # RepositorySpec(namespace='some_editable_namespace', name='some_checkout',
    #                version=None, fetch_method='EDITABLE', scm=None,
    #                spec_string='/tmp/pytest-of-adrian/pytest-79/test_editable_fetch_fetch0/some_checkout',
    #                src='/tmp/pytest-of-adrian/pytest-79/test_editable_fetch_fetch0/some_checkout')
    log.debug('req_spec: %r', req_spec)

    fetcher = editable.EditableFetch(galaxy_context, req_spec)

    find_results = {'custom': {'real_path': dest_tmp_path.strpath},
                    'content': []}

    res = fetcher.fetch(find_results=find_results)
    log.debug('res: %s', res)

    expected_link_name = os.path.join(galaxy_context.content_path,
                                      namespace_override,
                                      name)
    log.debug('expected_link_name: %s', expected_link_name)

    assert isinstance(res, dict)
    assert res['archive_path'] == dest_tmp_path.strpath
    assert res['custom']['symlinked_repo_root'] == expected_link_name




from ansible_galaxy.exceptions import GalaxyError


class GalaxyCliError(GalaxyError):
    pass


# replacement for AnsibleOptionError
# FIXME: CliOptionError (singular Option) ?
class CliOptionsError(Exception):
    pass



import logging
import os

import attr

from ansible_galaxy.models.repository_spec import RepositorySpec

log = logging.getLogger(__name__)


@attr.s(frozen=True)
class InstallDestinationInfo(object):
    destination_root_dir = attr.ib()
    repository_spec = attr.ib(type=RepositorySpec)

    # NOTE: destination_root_dir is the root of where the install begins
    #       but extract_archive_to is where the archive is extract to.
    #       The normal case (collection or multi-content archives) these values
    #       are the same (for ex, ~/.ansible/content/alikins/some_collection), but
    #       for trad roles they are different (for ex,
    #       destination_root_dir=~/.ansible/content/geerlingguy/ntp when
    #       extract_archive_to_dir=~/.ansible/content/geerlingguy/ntp/roles/ntp)
    extract_archive_to_dir = attr.ib()

    # 'alikins/collections_reqs_test/'
    namespaced_repository_path = attr.ib()

    install_info_path = attr.ib()

    force_overwrite = attr.ib(default=False)

    editable = attr.ib(default=False)

    @property
    def path(self):
        '''The full path to the eventually installed repository

        For, /home/user/.ansible/content/geerlingguy/ntp
        '''
        return os.path.join(self.destination_root_dir,
                            self.namespaced_repository_path)



import logging

from ansible_galaxy.config import yaml_file
from ansible_galaxy import exceptions

log = logging.getLogger(__name__)


# TODO: cache?  ideally, should only ever call this once per process
def load(full_file_path):
    try:
        with open(full_file_path, 'r') as config_file_stream:
            return yaml_file.load(config_file_stream)
    except (OSError, IOError) as e:
        log.exception(e)
        log.error('Unable to load config file (%s): %s', e.filename, e)
    except Exception as e:
        log.debug(e, exc_info=True)
        log.error('Unable to load and parse config file (%s): %s', full_file_path, e)
        # TODO: add a py2/py3 compat raise_from method for exception chaining
        raise exceptions.GalaxyConfigFileError(e, config_file_path=full_file_path)

    log.warning('No config data was loaded from file (%s), returning None instead', full_file_path)
    # TODO: empty dict instead?
    return None


def save(config_data, full_file_path):
    try:
        with open(full_file_path, 'w+') as config_file_stream:
            return yaml_file.save(config_data, config_file_stream)
    except Exception as e:
        log.exception(e)
        log.error('Unable to save config file (%s): %s', full_file_path, e)

        # just raise the exception again for now, may want to make this a
        raise



import logging

import attr

log = logging.getLogger(__name__)


@attr.s(frozen=True)
class MigrateRoleContext(object):
    role_path = attr.ib(validator=attr.validators.instance_of(str))
    output_path = attr.ib(validator=attr.validators.instance_of(str))
    output_force = attr.ib(default=False, validator=attr.validators.instance_of(bool))

    # roles always have a name, but not a namespace
    role_name = attr.ib(default=None)

    # collections always have namespace and name
    collection_namespace = attr.ib(default=None)
    collection_name = attr.ib(default=None)
    collection_version = attr.ib(default=None)
    # Note: Not validating this as valid spdx here, collection_info
    #       will do that later
    collection_license = attr.ib(default=None)



from collections import OrderedDict

import errno
import logging
import os
import shutil

from ansible_galaxy import exceptions
from ansible_galaxy import role_metadata
from ansible_galaxy.data import spdx_licenses
from ansible_galaxy.models.collection_info import CollectionInfo
from ansible_galaxy import yaml_persist

log = logging.getLogger(__name__)

# TODO: split this up into small methods


def migrate(migrate_role_context,
            display_callback):

    log.debug('migrate_role_context: %s', migrate_role_context)

    # validate paths

    # load role from migrate_role_context.role_path
    # though, that may be just loading the role_path/meta/main.yml

    dir_basename = os.path.basename(os.path.normpath(migrate_role_context.role_path))
    log.debug('dir_basename: %s', dir_basename)

    name_parts = dir_basename.split('.', 1)

    role_name = name_parts[0]
    role_namespace = None

    if len(name_parts) == 2:
        role_namespace = name_parts[0]
        role_name = name_parts[1]

    role_name = migrate_role_context.role_name or role_name

    collection_name = migrate_role_context.collection_name or role_name
    collection_namespace = migrate_role_context.collection_namespace or role_namespace

    role_md = role_metadata.load_from_dir(migrate_role_context.role_path,
                                          role_name=migrate_role_context.role_name)

    log.debug('role_metadata: %s', role_md)

    valid_licenses = spdx_licenses.get_spdx()

    # Check this here so errors can show more context
    if valid_licenses.get(role_md.license, None) is None:
        display_callback('The value for "license" found in "%s" is "%s" which is not a valid SPDX license' %
                         (migrate_role_context.role_path, role_md.license),
                         level='warning')

    if valid_licenses.get(migrate_role_context.collection_license, None) is None:
        display_callback('The value provided by --license is "%s" which is not a valid SPDX license' %
                         migrate_role_context.collection_license,
                         level='warning')

    collection_license = migrate_role_context.collection_license or role_md.license

    #  maybe some file tree walking

    # (or just let CollectionInfo construct fail...)
    # verify/validate namespace and names are valid for collections
    # verify version number is valid for collection

    # create/populate dicts for collection_info
    # fill in namespace, name, version, deps, tags, etc

    # migrate role style dependencies dict to a collections style
    # list of dicts
    # collection_deps = OrderedDict([tuple(key, role_md.dependencies[key]) for key in role_md.dependencies])
    collection_deps = OrderedDict([(req.requirement_spec.label, str(req.requirement_spec.version_spec)) for req in role_md.dependencies])

    log.debug('collection_deps: %s', collection_deps)

    # FIXME: license format is different, will have to map
    collection_info_dict = OrderedDict([
        # TODO: namespace, name
        ('namespace', collection_namespace),
        ('name', collection_name),
        ('version', migrate_role_context.collection_version),
        ('license', collection_license),
        ('description', role_md.description),
        # FIXME: verify role_md.author is a list or single
        ('authors', [role_md.author]),
        ('tags', role_md.galaxy_tags),
        ('issues', role_md.issue_tracker),
        ('dependencies', collection_deps)
    ])

    log.debug('collection_info_dict: %s', collection_info_dict)

    # create a CollectionInfo
    try:
        col_info = CollectionInfo(**collection_info_dict)
    except ValueError as e:
        raise exceptions.GalaxyClientError(e)

    log.debug('col_info: %s', col_info)

    # write out galaxy.yml
    # persist CollectionInfo to output_path/galaxy.yml
    output_filename = os.path.join(migrate_role_context.output_path, 'galaxy.yml')
    yaml_persist.save(col_info, output_filename)

    # hmmm... should probably have a Collection/Repository save()
    # support, but if not, do the equiv

    # create any needed dirs in output_path/ like roles/
    output_roles_dirpath = os.path.join(migrate_role_context.output_path, 'roles')
    log.debug('creating dir at %s', output_roles_dirpath)

    try:
        os.makedirs(output_roles_dirpath)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    #  the roles/ subdir for this role name
    named_role_subdir_in_output = os.path.join(output_roles_dirpath, collection_name)

    # If the dest roles/$rolename/ dir exists, and --force was used, then delete roles/$rolename/ so we
    # can recreate it. It that dir exists and --force isn't provided, there will be an exception raised
    if migrate_role_context.output_force:
        log.debug('removing %s', named_role_subdir_in_output)
        shutil.rmtree(named_role_subdir_in_output)

    # TODO: should we migrate plugins and modules to be collection level
    #       or just keep the role level?
    log.debug('copytree %s -> %s', migrate_role_context.role_path, named_role_subdir_in_output)

    # cp role_path/role_stuff* dirs to output_path
    # FIXME/NOTE: This does nothing clever, it doesn't ignore anything, it doesn't copy an
    #             explicit list of dirs, doesn't ignore .git, etc.
    shutil.copytree(migrate_role_context.role_path, named_role_subdir_in_output)

    # TODO: if there are modules or plugins using a bundled module_utils
    #       just moving the plugins isnt enough since the source code itself
    #       will need to be updated to use new plugin loader style paths

    # display any collected errors or messages
    # display the output path, maybe some summary of the migration results

    return os.EX_OK  # 0:


def _migrate(migrate_role_context, display_callback):
    pass




from ansible_galaxy.utils import repository_name

import logging
log = logging.getLogger(__name__)


def test_parse_content_name():
    _content_name = 'alikins.testing-content'

    user_name, repo_name, _content_name = repository_name.parse_repository_name(_content_name)

    assert user_name == 'alikins'
    assert _content_name == 'testing-content'
    assert repo_name is None

    _content_name = 'alikins.testing-content.elasticsearch_plugin.py'
    user_name, repo_name, _content_name = repository_name.parse_repository_name(_content_name)

    assert user_name == 'alikins'
    assert _content_name == 'elasticsearch_plugin.py'
    assert repo_name == 'testing-content'

    # TODO: revisit rules for content name /spec, ie, can
    #       github user ids or namespaces have '.' in them,
    #       and/or can specs for specific contents (a specific module
    #       in a repo) can have a '.' like 'eleastic_search.py'
    # content_name = 'somedotuser.dotuser.testing-content'
    # user_name, repo_name, content_name = content.parse_content_name(content_name)

    # assert user_name == 'somedotuser.dotuser'
    # assert content_name == 'testing-content'




import logging

from ansible_galaxy import display
from ansible_galaxy import matchers
from ansible_galaxy import installed_repository_db
from ansible_galaxy.utils.repository_name import parse_repository_name
from ansible_galaxy.utils.text import to_text

log = logging.getLogger(__name__)

SKIP_INFO_KEYS = ("name", "description", "readme_html", "related", "summary_fields", "average_aw_composite", "average_aw_score", "url")

GALAXY_REPOSITORY_TEMPLATE = """
label: {repo_label}
namespace: {repo_namespace}
name: {repo_name}
format: {repo_format}
latest_version: {repo_latest_version}
description: {repo_description}
scm: {repo_clone_url}

{repo_contents}"""

INSTALLED_REPOSITORY_TEMPLATE = """namespace: {repository.repository_spec.namespace}
name: {repository.repository_spec.name}
version: {repository.repository_spec.version}
path: {repository.path}
scm: {repository.repository_spec.scm}
"""


CONTENT_ITEM_TEMPLATE = """    content_name: {content_name}
    content_type: {content_type}
    description: {content_description}
"""


# FIXME: format?
def _repr_remote_repo(remote_data):
    # log.debug('remote_data: %s', remote_data)

    # flatten some info for format simplicity
    repo_namespace = remote_data['summary_fields']['namespace']['name']
    repo_name = remote_data['name']
    remote_data['repo_label'] = "{namespace}.{name}".format(namespace=repo_namespace, name=repo_name)
    remote_data['repo_namespace'] = remote_data['summary_fields']['namespace']['name']
    remote_data['repo_description'] = remote_data['description']
    remote_data['repo_clone_url'] = remote_data['clone_url']
    # remote_data['repo_name'] = remote_data['summary_fields']['repository']['original_name']
    remote_data['repo_name'] = remote_data['name']
    remote_data['repo_format'] = remote_data['format']

    content_info_parts = []
    content_item_objects = remote_data['summary_fields']['content_objects']

    for content_item_object in content_item_objects:
        content_item_data = {}
        content_item_data['content_name'] = content_item_object['name']
        content_item_data['content_type'] = content_item_object['content_type']
        content_item_data['content_description'] = content_item_object['description']

        content_str = CONTENT_ITEM_TEMPLATE.format(**content_item_data)
        # log.debug('content_str: %s', content_str)
        content_info_parts.append(content_str)

    versions = remote_data['summary_fields']['versions']
    try:
        latest_version = versions[0].get('version', 'N/A')
    except IndexError:
        latest_version = ""

    remote_data['repo_latest_version'] = latest_version

    remote_data['repo_contents'] = '\n'.join(content_info_parts)

    return GALAXY_REPOSITORY_TEMPLATE.format(**remote_data)


def _repr_installed_repository(installed_repository):
    return INSTALLED_REPOSITORY_TEMPLATE.format(repository=installed_repository)


def _repr_unmatched_label(unmatched_label):
    return '{0}'.format(unmatched_label)


# TODO: move to a repr for Role?
def _repr_role_info(role_info):

    text = [u"", u"Role: %s" % to_text(role_info['name'])]
    text.append(u"\tdescription: %s" % role_info.get('description', ''))

    for k in sorted(role_info.keys()):

        if k in SKIP_INFO_KEYS:
            continue

        if isinstance(role_info[k], dict):
            text.append(u"\t%s:" % (k))
            for key in sorted(role_info[k].keys()):
                if key in SKIP_INFO_KEYS:
                    continue
                text.append(u"\t\t%s: %s" % (key, role_info[k][key]))
        else:
            text.append(u"\t%s: %s" % (k, role_info[k]))

    return u'\n'.join(text)


def info_repository_specs(galaxy_context,
                          api,
                          repository_spec_strings,
                          display_callback=None,
                          offline=None):

    online = not offline

    display_callback = display_callback or display.display_callback

    offline = offline or False

    irdb = installed_repository_db.InstalledRepositoryDatabase(galaxy_context)

    labels_to_match = []

    all_labels_to_match = []
    for repository_spec_string in repository_spec_strings:
        galaxy_namespace, repository_name, content_name = parse_repository_name(repository_spec_string)

        log.debug('showing info for repository spec: %s', repository_spec_string)

        repository_name = repository_name or content_name

        if online:
            remote_data = api.lookup_repo_by_name(galaxy_namespace, repository_name)
            if remote_data:
                display_callback(_repr_remote_repo(remote_data))

        label_to_match = '%s.%s' % (galaxy_namespace, repository_name)
        all_labels_to_match.append(label_to_match)

        labels_to_match.append(label_to_match)

    matcher = matchers.MatchRepositorySpec([label_and_spec[1] for label_and_spec in labels_to_match])
    matcher = matchers.MatchLabels(labels_to_match)

    matched_repositories = irdb.select(repository_spec_match_filter=matcher)

    remote_data = False

    matched_labels = []
    for matched_repository in matched_repositories:
        display_callback(_repr_installed_repository(matched_repository))
        matched_labels.append(matched_repository.repository_spec.label)

    unmatched_labels = set(all_labels_to_match).difference(set(matched_labels))

    if unmatched_labels:
        display_callback('These repositories were not found:')

        for unmatched_label in sorted(unmatched_labels):
            display_callback(_repr_unmatched_label(unmatched_label))

    return



import logging

from ansible_galaxy.fetch import base
# mv details of this here
from ansible_galaxy.utils import scm_archive


log = logging.getLogger(__name__)


class ScmUrlFetch(base.BaseFetch):
    fetch_method = 'scm_url'

    def __init__(self, requirement_spec):
        super(ScmUrlFetch, self).__init__()

        self.requirement_spec = requirement_spec

        self.remote_resource = requirement_spec.src

    # scm
    def find(self):
        results = {'content': {'galaxy_namespace': self.requirement_spec.namespace,
                               'repo_name': self.requirement_spec.name},
                   }
        results['custom'] = {'scm_url': self.requirement_spec.src}

        return results

    def fetch(self, find_results=None):
        find_results = find_results or {}
        repository_archive_path = scm_archive.scm_archive_content(src=self.requirement_spec.src,
                                                                  scm=self.requirement_spec.scm,
                                                                  name=self.requirement_spec.name,
                                                                  version=str(self.requirement_spec.version))
        self.local_path = repository_archive_path

        log.debug('repository_archive_path=%s', repository_archive_path)

        results = {'archive_path': repository_archive_path,
                   'download_url': self.requirement_spec.src,
                   'fetch_method': self.fetch_method}
        results['content'] = find_results['content']
        results['custom'] = find_results.get('custom', {})
        results['custom']['scm_url'] = self.requirement_spec.src

        # TODO: what the heck should the version be for a scm_url if one wasnt specified?
        return results



import logging

import attr

from ansible_galaxy.models.collection_artifact_file import CollectionArtifactFile

log = logging.getLogger(__name__)


def convert_list_to_artifact_file_list(val):
    '''Convert a list of dicts with file info into list of CollectionArtifactFile'''

    new_list = []
    for file_item in val:
        if isinstance(file_item, CollectionArtifactFile):
            new_list.append(file_item)
        else:
            artifact_file = CollectionArtifactFile(name=file_item['name'],
                                                   ftype=file_item['ftype'],
                                                   chksum_type=file_item.get('chksum_type'),
                                                   chksum_sha256=file_item.get('chksum_sha256'))
            new_list.append(artifact_file)

    return new_list


@attr.s(frozen=True)
class CollectionArtifactFileManifest(object):
    files = attr.ib(factory=list, converter=convert_list_to_artifact_file_list)

    format = attr.ib(default=1,
                     validator=attr.validators.instance_of(int))



import logging

import pytest

from ansible_galaxy_cli import cli

log = logging.getLogger(__name__)


def test_cli():
    with pytest.raises(TypeError, match="Can't instantiate abstract class CLI with abstract methods parse"):
        cli.CLI(args=[])
    # log.debug(excinfo.getrepr())



import logging

log = logging.getLogger(__name__)


# TODO: abc?
class Match(object):
    '''A callable condition/match

    self.match(other) should return a bool'''

    def __call__(self, other):
        return self.match(other)

    def match(self, other):
        return True


class MatchAll(Match):
    def match(self, other):
        return True


class MatchNone(Match):
    def match(self, other):
        return False


class MatchNames(Match):
    def __init__(self, names):
        self.names = names

    def match(self, other):
        return other.repository_spec.name in self.names


class MatchLabels(Match):
    def __init__(self, labels):
        self.labels = labels

    def match(self, other):
        return other.repository_spec.label in self.labels


class MatchRepositoryToRequirementSpec(Match):
    def __init__(self, requirement_specs):
        self.requirement_specs = requirement_specs

    def match(self, other):
        for req_spec in self.requirement_specs:
            if other.repository_spec.namespace != req_spec.namespace or \
                    other.repository_spec.name != req_spec.name or \
                    not req_spec.version_spec.match(other.repository_spec.version):
                log.debug('other: %s did not match %s', other, req_spec)
                return False
        return True
        # return other.requirement_spec in self.requirement_specs


class MatchRepositorySpec(Match):
    def __init__(self, repository_specs):
        self.repository_specs = repository_specs or []

    def match(self, other):
        res = other.repository_spec in self.repository_specs
        return res


class MatchRepositorySpecNamespaceName(Match):
    def __init__(self, repository_specs):
        self.namespaces_names = [(x.namespace, x.name) for x in repository_specs] or []

    def match(self, other):
        ns_n_match_res = False
        other_ns_n = (other.repository_spec.namespace, other.repository_spec.name)
        ns_n_match_res = other_ns_n in self.namespaces_names

        return ns_n_match_res


class MatchRepositorySpecNamespaceNameVersion(Match):
    def __init__(self, repository_specs):
        self.namespaces_names_versions = [(x.namespace, x.name, x.version) for x in repository_specs] or []
        self.namespaces_names = [(x.namespace, x.name) for x in repository_specs] or []

    def match(self, other):
        ns_n_v_match_res = False

        other_ns_n_v = (other.repository_spec.namespace, other.repository_spec.name, other.repository_spec.version)

        ns_n_v_match_res = other_ns_n_v in self.namespaces_names_versions

        return ns_n_v_match_res


class MatchNamespace(Match):
    def __init__(self, namespaces_to_match):
        self.namespaces_to_match = namespaces_to_match or []

    def match(self, other):
        res = other.namespace in self.namespaces_to_match
        return res


class MatchNamespacesOrLabels(Match):
    def __init__(self, namespaces_or_labels):
        self.namespaces_or_labels = namespaces_or_labels or []

    def match(self, other):
        # TODO: should this matcher require a namespace string or a namespace object? either?
        return any([other.repository_spec.label in self.namespaces_or_labels,
                    other.repository_spec.namespace in self.namespaces_or_labels])



import logging

import yaml

from ansible_galaxy.models.collection_info import \
    CollectionInfo
from ansible_galaxy import exceptions

log = logging.getLogger(__name__)

# DEFAULT_FILENAME = "collection_info.yml"
COLLECTION_INFO_FILENAME = "galaxy.yml"


# TODO: replace with a generic version for cases
#       where SomeClass(**dict_from_yaml) works
def load(data_or_file_object, klass=None):
    data_dict = yaml.safe_load(data_or_file_object)

    try:
        collection_info = CollectionInfo(**data_dict)
    except ValueError:
        raise
    except Exception as exc:
        raise exceptions.GalaxyClientError("Error parsing collection metadata: %s" % str(exc))

    return collection_info



import logging

from ansible_galaxy import exceptions

log = logging.getLogger(__name__)


def resolve(spec_data):
    # build up the 'name.namespace' spec based
    # split the name on '.' and recombine the first 1 or 2
    src = spec_data['src']
    name_parts = src.split('.')

    # enforce the galaxy content spec specific rule about requiring a dot in namespace.name

    # if we have namespace.name, pop off namespace and use rest for name
    if len(name_parts) < 2:
        raise exceptions.GalaxyError('A galaxy content spec must have at least a namespace, a dot, and a name but "%s" does not.'
                                     % spec_data.get('spec_string'))

    if len(name_parts) > 1:
        spec_data['namespace'] = name_parts[0]

        # use the second part of namespace.name if there wasnt an explicit name=foo
        if not spec_data['name']:
            spec_data['name'] = name_parts[1]
    else:
        spec_data['name'] = name_parts[0]

    return spec_data



########################################################################
#
# (C) 2013, James Cammarata <jcammarata@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import codecs
import logging
import json
import os
import socket
import ssl
import sys
import uuid

from six.moves.urllib.error import HTTPError
from six.moves.urllib.parse import quote as urlquote
from six.moves.urllib.parse import urlparse
from six.moves import http_client

from ansible_galaxy import __version__ as mazer_version
from ansible_galaxy import exceptions
from ansible_galaxy.multipart_form import MultiPartForm
from ansible_galaxy.utils.text import to_native, to_text

from ansible_galaxy.flat_rest_api.urls import open_url

log = logging.getLogger(__name__)
http_log = logging.getLogger('%s.(http).(general)' % __name__)
request_log = logging.getLogger('%s.(http).(request)' % __name__)
response_log = logging.getLogger('%s.(http).(response)' % __name__)

USER_AGENT_FORMAT = 'Mazer/{version} ({platform}; python:{py_major}.{py_minor}.{py_micro}) ansible_galaxy/{version}'


def user_agent():
    user_agent_data = {'version': mazer_version,
                       'platform': sys.platform,
                       'py_major': sys.version_info.major,
                       'py_minor': sys.version_info.minor,
                       'py_micro': sys.version_info.micro}
    return USER_AGENT_FORMAT.format(**user_agent_data)


def g_connect(method):
    ''' wrapper to lazily initialize connection info to galaxy '''

    def wrapped(self, *args, **kwargs):
        if not self.initialized:
            log.debug("Initial connection to galaxy_server: %s", self._api_server)

            server_version = self._get_server_api_version()

            if server_version not in self.SUPPORTED_VERSIONS:
                raise exceptions.GalaxyClientError("Unsupported Galaxy server API version: %s" % server_version)

            self.baseurl = '%s/api/%s' % (self._api_server, server_version)
            self.version = server_version  # for future use

            # log.debug("Base API: %s", self.baseurl)

            self.initialized = True
        return method(self, *args, **kwargs)
    return wrapped


class GalaxyAPI(object):
    ''' This class is meant to be used as a API client for an Ansible Galaxy server '''

    SUPPORTED_VERSIONS = ['v1', 'v2']

    # FIXME: just pass in server_url
    def __init__(self, galaxy):
        self.galaxy = galaxy

        # log.debug('galaxy: %s', galaxy)
        log.debug('Using galaxy server URL %s with ignore_certs=%s', galaxy.server['url'], galaxy.server['ignore_certs'])

        self._validate_certs = not galaxy.server['ignore_certs']
        self.baseurl = None
        self.version = None
        self.initialized = False
        self.log = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        # set the API server
        self._api_server = galaxy.server['url']

        # self.log.debug('Validate TLS certificates for %s: %s', self._api_server, self._validate_certs)

        self.user_agent = user_agent()
        log.debug('User Agent: %s', self.user_agent)

    # TODO: raise an API/net specific exception?
    @g_connect
    def __call_galaxy(self, url, args=None, headers=None, http_method=None):
        http_method = http_method or 'GET'
        headers = headers or {}
        request_id = uuid.uuid4().hex
        headers['X-Request-ID'] = request_id

        # The slug we use to identify a request by method, url and request id
        # For ex, '"GET https://galaxy.ansible.com/api/v1/repositories" c48937f4e8e849828772c4a0ce0fd5ed'
        request_slug = '"%s %s" %s' % (http_method, url, request_id)

        try:
            # log the http request_slug with request_id to the main log and
            # to the http log, both at INFO level for now.
            http_log.info('%s', request_slug)
            self.log.info('%s', request_slug)

            request_log.debug('%s args=%s', request_slug, args)
            request_log.debug('%s headers=%s', request_slug, headers)

            resp = open_url(url, data=args, validate_certs=self._validate_certs,
                            headers=headers, method=http_method,
                            http_agent=self.user_agent,
                            timeout=20)

            response_log.info('%s http_status=%s', request_slug, resp.getcode())

            final_url = resp.geturl()
            if final_url != url:
                request_log.debug('%s Redirected to: %s', request_slug, resp.geturl())

            resp_info = resp.info()
            response_log.debug('%s info:\n%s', request_slug, resp_info)

            # FIXME: making the request and loading the response should be sep try/except blocks
            response_body = to_text(resp.read(), errors='surrogate_or_strict')

            # debug log the raw response body
            response_log.debug('%s response body:\n%s', request_slug, response_body)

            data = json.loads(response_body)

            # debug log a json version of the data that was created from the response
            response_log.debug('%s data:\n%s', request_slug, json.dumps(data, indent=2))
        except HTTPError as http_exc:
            self.log.debug('Exception on %s', request_slug)
            self.log.exception("%s: %s", request_slug, http_exc)

            # FIXME: probably need a try/except here if the response body isnt json which
            #        can happen if a proxy mangles the response
            res = json.loads(to_text(http_exc.fp.read(), errors='surrogate_or_strict'))

            http_log.error('%s data from server error response:\n%s', request_slug, res)

            try:
                error_msg = 'HTTP error on request %s: %s' % (request_slug, res['detail'])
                raise exceptions.GalaxyClientError(error_msg)
            except (KeyError, TypeError) as detail_parse_exc:
                self.log.exception("%s: %s", request_slug, detail_parse_exc)
                self.log.warning('Unable to parse error detail from response for request: %s response:  %s', request_slug, detail_parse_exc)

            # TODO: great place to be able to use 'raise from'
            # FIXME: this needs to be tweaked so the
            raise exceptions.GalaxyClientError(http_exc)
        except (ssl.SSLError, socket.error) as e:
            self.log.debug('Connection error to Galaxy API for request %s: %s', request_slug, e)
            self.log.exception("%s: %s", request_slug, e)
            raise exceptions.GalaxyClientAPIConnectionError('Connection error to Galaxy API for request %s: %s' % (request_slug, e))

        return data

    @property
    def api_server(self):
        return self._api_server

    @property
    def validate_certs(self):
        return self._validate_certs

    def _get_server_api_version(self):
        """
        Fetches the Galaxy API current version to ensure
        the API server is up and reachable.
        """
        url = '%s/api/' % self._api_server

        try:
            return_data = open_url(url, validate_certs=self._validate_certs)
        except Exception as e:
            raise exceptions.GalaxyClientError("Failed to get data from the API server (%s): %s " % (url, to_native(e)))

        try:
            data = json.loads(to_text(return_data.read(), errors='surrogate_or_strict'))
        except Exception as e:
            raise exceptions.GalaxyClientError("Could not process data from the API server (%s): %s " % (url, to_native(e)))

        if 'current_version' not in data:
            raise exceptions.GalaxyClientError("missing required 'current_version' from server response (%s)" % url)

        self.log.debug('Server API version of URL %s is "%s"', url, data['current_version'])
        return data['current_version']

    @g_connect
    def lookup_repo_by_name(self, namespace, name):
        namespace = urlquote(namespace)
        name = urlquote(name)
        url = '%s/repositories/?name=%s&provider_namespace__namespace__name=%s' % (self.baseurl, name, namespace)
        data = self.__call_galaxy(url, http_method='GET')
        if data["results"]:
            return data["results"][0]
        return {}

    @g_connect
    def fetch_content_related(self, related_url):
        """
        Fetch the list of related items for the given role.
        The url comes from the 'related' field of the role.
        """
        self.log.debug('related_url=%s', related_url)

        # try:
        url = '%s%s?page_size=50' % (self._api_server, related_url)

        # can raise a GalaxyClientError
        data = self.__call_galaxy(url, http_method='GET')

        # empty list for return value if there are no results
        results = data.get('results', [])

        # TODO: generalize the pagination support
        # check for paginated results
        done = (data.get('next_link', None) is None)

        while not done:
            url = '%s%s' % (self._api_server, data['next_link'])
            data = self.__call_galaxy(url, http_method='GET')

            # if no results, default to a empty list
            results += data.get('results', [])

            done = (data.get('next_link', None) is None)

        return results

    @g_connect
    def fetch_namespace(self, namespace):
        namespace = urlquote(namespace)
        url = '%s/namespaces/?name=%s' % (self.baseurl, namespace)
        data = self.__call_galaxy(url, http_method='GET')
        if data["results"]:
            return data["results"][0]
        return {}

    @g_connect
    def publish_file(self, data, archive_path, publish_api_key):
        form = MultiPartForm()
        for key in data:
            form.add_field(key, data[key])

        form.add_file('file', os.path.basename(archive_path),
                      fileHandle=codecs.open(archive_path, "rb"),
                      mimetype='application/octet-stream')

        # FIXME: override api version in POST to collections/ as 'v2' for now, as
        #        self.baseurl will automatically choose v1 based on 'GET /api'
        #        results, but collections/ api endpoints are v2 only.
        # TODO: figure out how to track API versions finer grained? Ideally
        #       simple enough to not end up with adhoc HATEAOS imp
        #       Maybe just hardcode api ver in calls?
        collection_url_ver = 'v2'
        url = '%s/api/%s/collections/' % (self._api_server, collection_url_ver)

        log.debug('url: %s', url)

        _, netloc, url, _, _, _ = urlparse(url)

        try:
            form_buffer = form.get_binary().getvalue()
            http = http_client.HTTPConnection(netloc)
            http.connect()
            http.putrequest("POST", url)
            http.putheader('Content-type', form.get_content_type())
            http.putheader('Content-length', str(len(form_buffer)))

            if publish_api_key:
                http.putheader('Authorization', 'Token %s' % publish_api_key)

            http.endheaders()
            http.send(form_buffer)
        except socket.error as exc:
            log.exception(exc)
            raise exceptions.GalaxyPublishError(
                'Network error while transferring file "%s" to Galaxy server (%s): %s' %
                (archive_path, self.galaxy.server['url'], str(exc)),
                archive_path=archive_path
            )

        r = http.getresponse()

        log.debug('code: %s', r.getcode())
        log.debug('info: %s', r.info())
        log.debug('reason: %s', r.reason)

        response_body = r.read()
        log.debug('response_body: %s', response_body)

        # 202 'Accepted'
        if r.status == 202:
            return response_body
        else:
            raise exceptions.GalaxyPublishError(
                'Error transferring file "%s" to Galaxy server (%s): %s - %s' %
                (archive_path, self.galaxy.server['url'], r.status, r.reason),
                archive_path=archive_path
            )



import fnmatch
import logging
import os
import pprint

import attr

log = logging.getLogger(__name__)

# dirtools defaults to .git/.hg/.svn, but we need to extend that set
DEFAULT_IGNORE_DIRS = ['releases', 'CVS', '.bzr', '.hg', '.git', '.svn', '__pycache__',
                       '.tox']
DEFAULT_IGNORE_PATTERNS = ['*.pyc', '*.retry']

pf = pprint.pformat

# start with a ftw over collection_path, and refine from there


def file_is_excluded(filename, exclude_patterns):
    '''Check a filename against a list of exclude patterns'''
    return any([fnmatch.fnmatch(filename, pattern) for pattern in exclude_patterns])


@attr.s
class CollectionMember(object):
    '''The info need to add a file to a archive (orig path, dest path, etc)'''
    src_full_path = attr.ib()
    dest_relative_path = attr.ib()


@attr.s
class FileWalker(object):
    collection_path = attr.ib()
    file_errors = attr.ib(factory=list)
    ignore_dirs = attr.ib(default=DEFAULT_IGNORE_DIRS)
    exclude_patterns = attr.ib(default=DEFAULT_IGNORE_PATTERNS)

    def walk(self):
        full_collection_path = os.path.abspath(self.collection_path)

        yield full_collection_path

        for dirpath, dirnames, filenames in os.walk(full_collection_path,
                                                    onerror=self.on_walk_error,
                                                    followlinks=False):
            # filter out .git etc
            # NOTE: This modifies dirnames while it is being walked over
            for dirname in dirnames:
                if dirname in self.ignore_dirs:
                    # log.debug('ignoring dir: %s', dirname)
                    dirnames.remove(dirname)
                else:
                    dir_full_path = os.path.join(dirpath, dirname)
                    # log.debug('yield dir_full_path: %s', dir_full_path)
                    yield dir_full_path

            for filename in filenames:
                # log.debug('fn dirpath: %s filename: %s', dirpath, filename)

                full_path = os.path.join(dirpath, filename)
                # log.debug('full_path: %s', full_path)

                # collection_relative_path = os.path.relpath(full_path, self.collection_path)

                # log.debug('collection_relative_path: %s', collection_relative_path)

                if file_is_excluded(full_path, self.exclude_patterns):
                    continue

                # log.debug('yield (fn full_path)%s', full_path)
                yield full_path
                # yield collection_relative_path
                # yield CollectionMember(src_full_path=full_path, dest_relative_path=collection_relative_path)

    def relative_walk(self):
        for full_path in self.walk():
            # rel_path = collection_member.dest_relative_path
            relative_path = os.path.relpath(full_path, self.collection_path)

            yield relative_path

    def on_walk_error(self, walk_error):
        log.warning('walk error on %s: %s', walk_error.filename, walk_error)
        self.file_errors.append(walk_error)


@attr.s
class CollectionMembers(object):
    walker = attr.ib(default=None)

    def walk(self):
        return self.walker.walk()

    # TODO: could/should make CollectionMembers iterable with a iter() ?
    def run(self):
        '''collect/walk, then filter found members, yield the members'''
        _members = self.walk()

        for _member in _members:
            yield _member



import io
import logging

from ansible_galaxy import role_metadata
from ansible_galaxy.models.role_metadata import RoleMetadata

log = logging.getLogger(__name__)

yaml_data1 = u'''
---
dependencies: []

galaxy_info:
  author: geerlingguy
  description: Apache 2.x for Linux.
  company: "Midwestern Mac, LLC"
  license: "license (BSD, MIT)"
  min_ansible_version: 2.2
  platforms:
    - name: EL
      versions:
        - all
    - name: Amazon
      versions:
        - all
    - name: Debian
      versions:
        - all
    - name: Ubuntu
      versions:
        - precise
        - raring
        - saucy
        - trusty
        - xenial
    - name: Suse
      versions:
        - all
    - name: Solaris
      versions:
        - 11.3
  galaxy_tags:
    - web
    - apache
    - webserver
    - html

allow_duplicates: yes
'''


def test_load():
    yaml_fo = io.StringIO(initial_value=yaml_data1)
    role_md = role_metadata.load(yaml_fo, role_name='apache')

    assert isinstance(role_md, RoleMetadata)

    assert role_md.name == 'apache'
    assert role_md.min_ansible_version == '2.2'
    assert isinstance(role_md.galaxy_tags, list)
    assert 'web' in role_md.galaxy_tags
    assert role_md.galaxy_tags == ['web', 'apache', 'webserver', 'html']
    assert isinstance(role_md.allow_duplicates, bool)


def test_load_string():
    role_md = role_metadata.load(yaml_data1, role_name='apache')

    log.debug('role_md: %s', role_md)

    assert isinstance(role_md, RoleMetadata)
    assert role_md.name == 'apache'
    assert role_md.min_ansible_version == '2.2'
    assert isinstance(role_md.galaxy_tags, list)
    assert 'web' in role_md.galaxy_tags
    assert role_md.galaxy_tags == ['web', 'apache', 'webserver', 'html']
    assert isinstance(role_md.allow_duplicates, bool)



import datetime
import io
import logging

import yaml

from six import string_types

from ansible_galaxy import install_info
from ansible_galaxy.models.install_info import InstallInfo

log = logging.getLogger(__name__)

yaml_data1 = u'''
install_date: Tue Jul 17 14:41:59 2018
install_date_iso: 2018-07-17 14:41:59.229716
version: 0.1.0
'''


def test_load():
    yaml_fo = io.StringIO(initial_value=yaml_data1)
    install_info_ = install_info.load(yaml_fo)

    log.debug('install_info_: %s', install_info_)

    assert isinstance(install_info_, InstallInfo)

    assert str(install_info_.version) == '0.1.0'
    assert isinstance(install_info_.install_date_iso, datetime.datetime)
    assert isinstance(install_info_.install_date, string_types)


def test_load_string():
    install_info_ = install_info.load(yaml_data1)

    log.debug('install_info_: %s', install_info_)

    assert isinstance(install_info_, InstallInfo)

    assert str(install_info_.version) == '0.1.0'
    assert isinstance(install_info_.install_date_iso, datetime.datetime)
    assert isinstance(install_info_.install_date, string_types)


def test_load_from_filename(tmpdir):
    temp_dir = tmpdir.mkdir('mazer_content_install_info_unit_test')
    temp_file = temp_dir.join('.galaxy_install_info')
    temp_file.write(yaml_data1)

    log.debug('temp_file.strpath: %s', temp_file.strpath)

    install_info_ = install_info.load_from_filename(temp_file.strpath)

    log.debug('install_info_: %s', install_info_)

    assert isinstance(install_info_, InstallInfo)

    assert str(install_info_.version) == '0.1.0'
    assert isinstance(install_info_.install_date_iso, datetime.datetime)
    assert isinstance(install_info_.install_date, string_types)


def test_save(tmpdir):
    install_datetime = datetime.datetime.utcnow()
    install_info_ = InstallInfo.from_version_date(version='4.5.6',
                                                  install_datetime=install_datetime)
    log.debug('install_info: %s', install_info_)

    temp_dir = tmpdir.mkdir('mazer_content_install_info_unit_test')
    temp_file = temp_dir.join('.galaxy_install_info')
    install_info.save(install_info_.to_dict_version_strings(),
                      temp_file.strpath)

    log.debug('tmpfile: %s', temp_file)

    res = temp_file.read()
    log.debug('res: %s', res)

    reloaded = yaml.safe_load(res)

    assert isinstance(reloaded, dict)
    assert reloaded['version'] == '4.5.6'
    assert reloaded['install_date_iso'] == install_datetime

# TODO: test perms, etc



import logging
import os
import fnmatch

from ansible_galaxy import installed_content_item_db
from ansible_galaxy import matchers

log = logging.getLogger(__name__)


def test_installed_content_db(galaxy_context):
    icd = installed_content_item_db.InstalledContentItemDatabase(galaxy_context)

    for x in icd.select():
        log.debug('x: %s', x)


def test_installed_content_db_match_names(galaxy_context):
    icd = installed_content_item_db.InstalledContentItemDatabase(galaxy_context)

    match_filter = matchers.MatchNames(['foo.bar'])
    for x in icd.select(content_match_filter=match_filter):
        log.debug('x: %s', x)


def test_installed_content_iterator(galaxy_context):
    ici = installed_content_item_db.installed_content_item_iterator(galaxy_context,
                                                                    content_item_type='role')

    for i in ici:
        log.debug(i)


class MatchContentItemNamesFnmatch(object):
    def __init__(self, fnmatch_patterns):
        self.fnmatch_patterns = fnmatch_patterns

    def __call__(self, other):
        return self.match(other)

    def match(self, other):
        # log.debug('self.fnmatch_patterns: %s other.name: %s', self.fnmatch_patterns, other.name)
        # log.debug('fnm: %s', fnmatch.fnmatch(other.name, self.fnmatch_patterns[0]))
        return any([fnmatch.fnmatch(other.name, x) for x in self.fnmatch_patterns])


def test_installed_content_item_iterator_empty(galaxy_context, mocker):
    mocker.patch('ansible_galaxy.installed_namespaces_db.get_namespace_paths',
                 return_value=iter(['foo', 'blip']))
    mocker.patch('ansible_galaxy.repository.os.path.isdir',
                 return_value=True)
    mocker.patch('ansible_galaxy.installed_repository_db.get_repository_paths',
                 return_value=iter(['bar', 'baz']))

    mocker.patch('ansible_galaxy.installed_content_item_db.glob_content_path_iterator',
                 return_value=iter(['/dev/null/content/foo/bar/roles/role-1',
                                    '/dev/null/content/blip/baz/roles/role-2']))

    ici = installed_content_item_db.installed_content_item_iterator(galaxy_context,
                                                                    content_item_type='roles')
    log.debug('ici: %s', ici)

    installed_content_item = list(ici)
    log.debug('installed_content_item: %s', installed_content_item)
    assert installed_content_item[0]['content_data'].name == 'role-1'


def test_installed_content_item_iterator_tmp_content(galaxy_context):
    ici = installed_content_item_db.installed_content_item_iterator(galaxy_context,
                                                                    content_item_type='roles')

    content_path = galaxy_context.content_path

    # make some 'namespace_paths' in temp content_path
    tmp_namespace_paths = ['ns1.repo1', 'ns1.repo2', 'ns2.repo1']
    content_item_type_dirs = ['roles']
    content_item_names = ['somerole1', 'somerole2', 'role3']
    role_subdirs = ['meta', 'tasks']

    # TODO/FIXME: replace file access/tmp stuff with mocking
    for tmp_namespace_path in tmp_namespace_paths:
        for content_item_type_dir in content_item_type_dirs:
            for content_item_name in content_item_names:
                for role_subdir in role_subdirs:
                    full_content_item_path = os.path.join(content_path, tmp_namespace_path, content_item_type_dir, content_item_name, role_subdir)
                    log.debug('full_content_path: %s', full_content_item_path)
                    os.makedirs(full_content_item_path)

                    main_yml = os.path.join(full_content_item_path, 'main.yml')
                    with open(main_yml, 'w') as fd:
                        fd.write('---\n')

                    if role_subdir == 'meta':
                        install_info = os.path.join(full_content_item_path, '.galaxy_install_info')
                        with open(install_info, 'w') as fd:
                            fd.write('version: 1.2.3\n')

    for i in ici:
        log.debug('ici stuff: %s', i)

    ici2 = \
        installed_content_item_db.installed_content_item_iterator(galaxy_context,
                                                                  content_item_match_filter=MatchContentItemNamesFnmatch(['somerole*', 'asdfsd']),
                                                                  content_item_type='roles')

    for i in ici2:
        log.debug('i2: %s', i)




import logging
import pprint

import pytest

from ansible_galaxy import exceptions
from ansible_galaxy import install
from ansible_galaxy.models.repository import Repository
from ansible_galaxy.models.repository_spec import RepositorySpec

log = logging.getLogger(__name__)
pf = pprint.pformat


def display_callback(msg, **kwargs):
    log.debug(msg)


def test_install_no_valid_content(galaxy_context, mocker):
    repo_spec = RepositorySpec(namespace='some_namespace',
                               name='some_name',
                               version='4.3.2')

    mock_fetcher = mocker.MagicMock(name='MockFetch')
    fetch_results = {}

    with pytest.raises(exceptions.GalaxyClientError, match='No valid content data found for') as exc_info:
        install.install(galaxy_context,
                        fetcher=mock_fetcher,
                        fetch_results=fetch_results,
                        repository_spec=repo_spec,
                        display_callback=display_callback)
    log.debug('exc_info: %s', exc_info)
    # log.debug('res: %s', res)


def test_install(galaxy_context, mocker):
    repo_spec = RepositorySpec(namespace='some_namespace',
                               name='some_name',
                               version='4.3.2')

    mock_fetcher = mocker.MagicMock(name='MockFetch')
    fetch_results = {'archive_path': '/dev/null/doesntexist'}

    # Mock args for creating a Mock to replace repository_archive.load_archive that returns
    # a Mock'ed RepositoryArchive with a mocked repository_dest_root_subpath() that returns ''
    config = {'return_value': mocker.MagicMock(name='MockRepoArchive', **{'repository_dest_root_subpath.return_value': ''})}

    mocker.patch.object(install.repository_archive, 'load_archive', **config)

    res = install.install(galaxy_context,
                          fetcher=mock_fetcher,
                          fetch_results=fetch_results,
                          repository_spec=repo_spec,
                          display_callback=display_callback)
    log.debug('res: %s', res)

    assert isinstance(res, list)
    assert len(res) > 0
    assert isinstance(res[0], Repository)
    assert res[0].repository_spec == repo_spec
    assert galaxy_context.content_path in res[0].path


def test_find(mocker):
    mock_fetcher = mocker.MagicMock(name='MockFetch')
    mock_fetcher.find.return_value = {}

    res = install.find(mock_fetcher)

    log.debug('res: %s', res)

    assert res == {}


def test_fetch(mocker):
    mock_fetcher = mocker.MagicMock(name='MockFetch')
    mock_fetcher.fetch.return_value = {}
    repo_spec = RepositorySpec(namespace='some_namespace',
                               name='some_name',
                               version='86.75.30')
    find_results = {}

    res = install.fetch(mock_fetcher, repo_spec, find_results)

    log.debug('res: %s', res)


def test_fetch_download_error(mocker):
    mock_fetcher = mocker.MagicMock(name='MockFetch')
    mock_fetcher.fetch.side_effect = exceptions.GalaxyDownloadError(url='http://example.invalid')

    repo_spec = RepositorySpec(namespace='some_namespace',
                               name='some_name',
                               version='86.75.30')
    find_results = {}

    with pytest.raises(exceptions.GalaxyDownloadError) as exc_info:
        install.fetch(mock_fetcher, repo_spec, find_results)
    log.debug("exc_info: %s", exc_info)



import logging
import os

from ansible_galaxy import collection_artifact_file_manifest
from ansible_galaxy import yaml_persist
from ansible_galaxy.models.collection_artifact_file import CollectionArtifactFile
from ansible_galaxy.models.collection_artifact_file_manifest import \
    CollectionArtifactFileManifest

log = logging.getLogger(__name__)


def test_load():
    file_name = "example_artifact_file_manifest1.yml"
    test_data_path = os.path.join(os.path.dirname(__file__), '%s' % file_name)

    with open(test_data_path, 'r') as data_fd:
        res = collection_artifact_file_manifest.load(data_fd)
        log.debug('res: %s', res)

        assert isinstance(res, CollectionArtifactFileManifest)
        assert isinstance(res.files, list)
        assert isinstance(res.files[0], CollectionArtifactFile)

        assert res.files[0].name == 'roles/some_role/defaults/main.yml'
        assert res.files[0].ftype == 'file'
        assert res.files[0].chksum_type == 'sha256'
        assert res.files[0].chksum_sha256 == 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'


def test_save(tmpdir):
    temp_dir = tmpdir.mkdir('mazer_collection_archive_file_manifest_unit_test')

    file_name = "example_artifact_file_manifest1.yml"

    temp_file = temp_dir.join(file_name)

    test_data_path = os.path.join(os.path.dirname(__file__), '%s' % file_name)
    with open(test_data_path, 'r') as data_fd:
        res = collection_artifact_file_manifest.load(data_fd)

        log.debug('temp_file.strpath: %s', temp_file.strpath)
        yaml_persist.save(res, temp_file.strpath)

        # open the save()'ed file and load a new collection_artifact_manifest from it
        with open(temp_file.strpath, 'r') as read_fd:
            reloaded = collection_artifact_file_manifest.load(read_fd)

            log.debug('reloaded: %s', reloaded)

            # verify the object loaded from known example matches the example after
            # a save()/load() cycle
            assert reloaded == res

            # read the file again and log the file contents
            read_fd.seek(0)
            buf = read_fd.read()
            log.debug('buf: %s', buf)



"""Exceptions used by ansible_galaxy client library"""


class GalaxyError(Exception):
    """Base exception for ansible_galaxy library and ansible_galaxy_cli app"""
    pass


class GalaxyClientError(GalaxyError):
    """Base exception for ansible_galaxy library

    Exceptions that escape from ansible_galaxy. should be based on this class"""
    pass


class ParserError(GalaxyError):
    """Base exception raised for errors while parsing galaxy content"""
    pass


# TODO: attrs for http code, url, msg or just reuse http exception from elsewhere
class GalaxyDownloadError(GalaxyError):
    '''Raise when there is an error downloading galaxy content'''

    def __init__(self, *args, **kwargs):
        url = kwargs.pop('url', None)
        super(GalaxyDownloadError, self).__init__(*args, **kwargs)
        self.url = url

    def __str__(self):
        url_blurb = ': '
        if self.url:
            url_blurb = ' (%s): ' % self.url
        msg = 'Error downloading%s%s' % (url_blurb, super(GalaxyDownloadError, self).__str__())
        return msg

    def __repr__(self):
        return '%s(url=%s, %s)' % (self.__class__.__name__, self.url, self.args)


class GalaxyRepositorySpecError(GalaxyClientError):
    '''Raised if the repository_spec was invalid'''
    def __init__(self, *args, **kwargs):
        repository_spec = kwargs.pop('repository_spec', None)
        super(GalaxyRepositorySpecError, self).__init__(*args, **kwargs)
        self.repository_spec = repository_spec


class GalaxyClientAPIConnectionError(GalaxyClientError):
    '''Raised if there were errors connecting to the Galaxy REST API'''
    pass


# TODO: proper rst docstrings with api info
class GalaxyConfigFileError(GalaxyClientError):
    '''Raised where there is an error loading or parsing a config file

       has a 'config_file_path' attribute with the config file path'''

    def __init__(self, *args, **kwargs):
        config_file_path = kwargs.pop('config_file_path', None)
        super(GalaxyConfigFileError, self).__init__(*args, **kwargs)
        self.config_file_path = config_file_path


class GalaxyArchiveError(GalaxyClientError):
    '''Raised for errors related to content archive (tar) files'''

    def __init__(self, *args, **kwargs):
        archive_path = kwargs.pop('archive_path', None)
        super(GalaxyArchiveError, self).__init__(*args, **kwargs)
        self.archive_path = archive_path


class GalaxyPublishError(GalaxyClientError):
    ''' Raised for errors related to publish command '''

    def __init__(self, msg, *args, **kwargs):
        self.msg = msg
        self.archive_path = kwargs.pop('archive_path', None)
        super(GalaxyPublishError, self).__init__(*args, **kwargs)

    def __str__(self, *args, **kwargs):
        msg = 'Error publishing %s - %s' % (self.archive_path, self.msg)
        return msg




import logging

import attr

# from ansible_galaxy.models.collection_info import CollectionInfo
log = logging.getLogger(__name__)


# see https://github.com/ansible/galaxy/issues/957
@attr.s(frozen=True)
class CollectionArtifactFile(object):
    name = attr.ib()
    ftype = attr.ib()
    src_name = attr.ib(default=None)
    chksum_type = attr.ib(default="sha256")
    chksum_sha256 = attr.ib(default=None)
    format = attr.ib(default=1)



import os


def get_config_path():
    paths = [
        'mazer.yml',
        '~/.ansible/mazer.yml',
        '/etc/ansible/mazer.yml'
    ]
    for path in paths:
        if os.path.exists(os.path.expanduser(path)):
            return path
    return paths[1]


# a list of tuples that is fed to an OrderedDict
DEFAULTS = [
    ('server',
     {'url': 'https://galaxy.ansible.com',
      'ignore_certs': False}
     ),

    # In order of priority
    ('content_path', '~/.ansible/collections/ansible_collections'),
    ('global_content_path', '/usr/share/ansible/collections/ansible_collections'),

    # runtime options
    ('options',
     {'role_skeleton_path': None,
      'role_skeleton_ignore': ["^.git$", "^.*/.git_keep$"]}
     ),
    ('version', 1),
]

# FIXME: replace with logging config
VERBOSITY = 0
CONFIG_FILE = get_config_path()



import logging

import pytest

from ansible_galaxy import exceptions
from ansible_galaxy.utils import scm_archive


log = logging.getLogger(__name__)

# TODO: Replace cloning from our own local non-bare git repo, with
#       init/populating a test repo in tmp. Ditto for hg.


def test_scm_archive_content_git(tmp_git_repo):
    res = scm_archive.scm_archive_content(src='file://%s' % tmp_git_repo.strpath,
                                          name='mazer_example',
                                          scm='git')

    log.debug('res: %s', res)


def test_scm_archive_content_git_branch_does_not_exist(tmp_git_repo):
    with pytest.raises(exceptions.GalaxyClientError,
                       match=r'- command git checkout .* failed in directory .* \(rc=.*\)') as exc_info:
        scm_archive.scm_archive_content(src='file://%s' % tmp_git_repo.strpath,
                                        name='mazer_example',
                                        scm='git',
                                        version='__branch_name_unlikely_to_exist_w7asyrf7say4fsyrdf__')

    log.debug('exc_info: %s', exc_info)


def test_scm_archive_content_git_scm_src_does_not_exist(tmp_git_repo):
    with pytest.raises(exceptions.GalaxyClientError, match=r'- command .* failed in directory .* \(rc=.*\)') as exc_info:
        scm_archive.scm_archive_content(src='file://%s/__this__should_not_exist__88489er7fw4f__' % tmp_git_repo.strpath,
                                        name='mazer_example',
                                        scm='git')

    # log.debug('res: %s', res)
    log.debug('exc_info: %s', exc_info)



import logging

from ansible_galaxy.models.requirement import Requirement, RequirementOps, RequirementScopes
from ansible_galaxy.models.requirement_spec import RequirementSpec
from ansible_galaxy.repository_spec import spec_data_from_string

log = logging.getLogger(__name__)


def from_dependency_spec_strings(dependency_spec_strings, namespace_override=None, editable=False):
    deps = []
    for dep_spec_string in dependency_spec_strings:
        dep_spec_data = spec_data_from_string(dep_spec_string)

        log.debug('dep_spec_data: %s', dep_spec_data)

        dep_spec = RequirementSpec.from_dict(dep_spec_data)

        log.debug('dep_spec: %s', dep_spec)

        # Add a requirement, but with the 'RUNTIME' scope
        requirement = Requirement(repository_spec=None, op=RequirementOps.EQ,
                                  scope=RequirementScopes.RUNTIME,
                                  requirement_spec=dep_spec)
        deps.append(requirement)

    return deps



import logging
import os
import sys

from ansible_galaxy import exceptions
from ansible_galaxy_cli.logger.setup import setup_default
from ansible_galaxy_cli.cli import galaxy
from ansible_galaxy_cli import exceptions as cli_exceptions

log = logging.getLogger(__name__)
stderr_log = logging.getLogger('%s.(stderr)' % __package__)


def main(args=None):
    setup_default()

    args = args or sys.argv[:]

    # import logging_tree
    # logging_tree.printout()

    log.debug('args: %s', args)

    cli = galaxy.GalaxyCLI(args)

    try:
        cli.parse()
    except cli_exceptions.CliOptionsError as e:
        cli.parser.print_help()
        log.error(e)
        return os.EX_USAGE

    # TODO: some level of exception mapper to set exit code based on exception
    try:
        exit_code = cli.run()
    except exceptions.GalaxyConfigFileError as e:
        log.exception(e)

        # The str(e)/error here maybe a multiline yaml error that looks
        # best on a fresh line
        stderr_log.error('Error loading configuration file %s:', e.config_file_path)
        stderr_log.error(e)

        return os.EX_CONFIG
    except exceptions.GalaxyError as e:
        log.exception(e)
        stderr_log.error(e)

        # exit with EX_SOFTWARE on generic error
        exit_code = os.EX_SOFTWARE
    except Exception as e:
        log.exception(e)
        stderr_log.error(e)
        exit_code = os.EX_SOFTWARE

        # let non-Galaxy exceptions bubble up and traceback
        log.debug('Uncaught exception for invocation: %s', cli._orig_args_copy)
        log.error('Uncaught exception, existing with exit code: %s', exit_code)
        raise

    log.debug('exit code: %s', exit_code)
    # do any return code setup we need here
    return exit_code



import logging

import attr

from ansible_galaxy.models.repository_spec import RepositorySpec
from ansible_galaxy.models.requirement_spec import RequirementSpec

log = logging.getLogger(__name__)

EXAMPLE = '''
# from galaxy
- src: yatesr.timezone

# from GitHub
- src: https://github.com/bennojoy/nginx

# from GitHub, overriding the name and specifying a specific tag
- src: https://github.com/bennojoy/nginx
  version: master
  name: nginx_role

# from a webserver, where the role is packaged in a tar.gz
# - src: https://some.webserver.example.com/files/master.tar.gz
#  name: http-role

# from Bitbucket
# - src: git+https://bitbucket.org/willthames/git-ansible-galaxy
#  version: v1.4

# from Bitbucket, alternative syntax and caveats
# - src: https://bitbucket.org/willthames/hg-ansible-galaxy
#  scm: hg

# from GitLab or other git-based scm, using git+ssh
# - src: git@gitlab.company.com:mygroup/ansible-base.git
#   scm: git
#  version: "0.1"  # quoted, so YAML doesn't parse this as a floating-point value
'''


# enum-ish, currently we only support exact match
# (and implicitly, 'any' via not setting it.
class RequirementOps(object):
    EQ = '='


class RequirementScopes(object):
    INSTALL = 'INSTALL'
    RUNTIME = 'RUNTIME'


@attr.s(frozen=True)
class Requirement(object):

    # The repo that is required. The RHS of the requirement.
    requirement_spec = attr.ib(type=RequirementSpec,
                               validator=attr.validators.instance_of(RequirementSpec))
    # requirement_spec = attr.ib(type=semanticVersion

    # the 'operation' or expression type of a requirement
    # for ex, an 'exact' match of namespace, name, version
    # or possibly a 'just name'
    # or >= a repository spec
    # or possible, some semver specific expression TBD
    #   (like 'major.minor == 1.2' or 'major' == 1 and 'minor' >= 3)
    #
    # At the moment, the support operations are:
    #  - exact namespace and name and full version
    #  - exact namespace and name (no version compare)
    #
    # Unclear if 'no version compare' should be annotated in
    # the RepositorySpec (ie, version=None) or in the Requirement op
    #
    # If 'no version compare' is up to the RepositorySpec, then the op here
    # would indicate '==', '>', '>=' etc
    #
    # TODO: figure out if we should go ahead and track an 'op' for each compontent
    #       of RepositorySpec. For example:
    # repo_spec_a(namespace='mynamespace', name='myname', version='1.0.0')
    #
    # if repo_spec_a needs req_repo_spec(namespace='some_req_namespace', name='some_req', version=None)
    # that could have a namespace_op ==, a name_op ==, and a version_op any? noop? *?
    #
    # With per field ops, we could support repo_spec_a needing
    # req_repo_spec(namespace=='*', name='some_req', version='1.2.3') so any namespace could provide
    # the req. Lots of ambiquity that way though...
    #
    # For misc notes and ideas see:
    # - https://golang.github.io/dep/docs/Gopkg.toml.html#version (a semver aware syntax for version matching)
    # - https://github.com/Masterminds/semver  - go lib for semver
    # - https://medium.com/@sdboyer/so-you-want-to-write-a-package-manager-4ae9c17d9527
    # - https://en.wikipedia.org/wiki/Conflict-driven_clause_learning  SAT solver variant
    # operation? constraint? comparison? predicate? expression?
    op = attr.ib(default=RequirementOps.EQ)

    # The repo that has a requirement, can be null
    repository_spec = attr.ib(default=None, type=RepositorySpec,
                              validator=attr.validators.optional(attr.validators.instance_of(RepositorySpec)))

    # do we need to resolve the dep at install time ('requirements') or
    # could we defer until runtime (like role dependencies)
    scope = attr.ib(default=RequirementScopes.INSTALL)

    def __str__(self):
        return '{repo_spec}->{req_spec_label}{op}{req_spec_version}'.format(repo_spec=str(self.repository_spec),
                                                                            req_spec_label=str(self.requirement_spec.label),
                                                                            op=self.op,
                                                                            req_spec_version=str(self.requirement_spec.version_spec),
                                                                            )

    # FIXME: just for debugging
    def __repr__(self):
        return str(self)



import logging
import re

import semantic_version

log = logging.getLogger(__name__)

VERSION_WITH_LEADING_V_MATCH_RE = re.compile(r'^[vV]\d+\.')
VERSION_WITH_LEADING_V_SUB_RE = re.compile(r'(^[vV])')


def convert_string_to_semver(version):
    # log.debug('vs: %s type: %s', version, type(version))

    if version is None:
        return None

    if isinstance(version, semantic_version.Version):
        return version

    return semantic_version.Version(version)


def convert_string_to_version_spec(version_spec):
    # TODO: or the default '*' spec?
    if version_spec is None:
        return None

    if isinstance(version_spec, semantic_version.Spec):
        return version_spec

    return semantic_version.Spec(version_spec)


def version_needs_aka(version_string):
    '''Return True if version_string is in v1.0.0 form'''
    if not version_string:
        return False

    matches = VERSION_WITH_LEADING_V_MATCH_RE.match(version_string)

    if not matches:
        return False

    return True


def normalize_version_string(version_string):
    '''Normalize any version strings (rm 'v' from 'v1.0.0' for ex


    https://github.com/ansible/galaxy-cli/wiki/Content-Versioning#versions-in-galaxy-cli

    "When providing a version, provide the semantic version with or without the leading 'v' or 'V'."

    strip off leading v or V, and return a version string without it.'''

    if not version_string:
        return version_string

    matches = VERSION_WITH_LEADING_V_MATCH_RE.match(version_string)

    if not matches:
        return version_string

    new_versions_string = VERSION_WITH_LEADING_V_SUB_RE.sub('', version_string, 1)

    log.warning('Stripping leading "v" or "V" from version string "%s", new version string is  "%s"',
                version_string, new_versions_string)

    return new_versions_string



import logging

from ansible_galaxy import exceptions
from ansible_galaxy.models.content import VALID_ROLE_SPEC_KEYS
from ansible_galaxy.utils import yaml_parse
from ansible_galaxy import repository_spec_parse
from ansible_galaxy import galaxy_repository_spec


log = logging.getLogger(__name__)


def parse_spec(repository_spec, resolver=None):
    result = yaml_parse.yaml_parse(repository_spec, resolver=resolver)
    log.debug('result=%s', result)
    return result


def assert_keys(repository_spec, name=None, version=None,
                scm=None, src=None, namespace=None):
    # name = name or ''
    # src = src or ''
    assert isinstance(repository_spec, dict)

    log.debug('repository_spec: %s', repository_spec)
    # TODO: should it default to empty string?
    assert repository_spec['name'] == name, \
        'repository_spec name=%s does not match expected name=%s' % (repository_spec['name'], name)
    assert repository_spec.get('namespace') == namespace
    assert repository_spec['version'] == version
    assert repository_spec['scm'] == scm
    assert repository_spec['src'] == src, \
        'repository_spec src=%s does not match expected src=%s' % (repository_spec['src'], src)


def test_yaml_parse_empty_string_no_namespace_required():
    spec = ''
    result = parse_spec(spec, resolver=repository_spec_parse.resolve)

    assert_keys(result, name='', version=None, scm=None, src='')


def test_yaml_parse_empty_string():
    spec = ''
    try:
        parse_spec(spec)
    except exceptions.GalaxyError as e:
        log.exception(e)
        return

    assert False, 'Expected a GalaxyError here because the content_spec didnt have a namespace'


def test_yaml_parse_just_name():
    spec = 'some_namespace.some_content'
    result = parse_spec(spec)

    assert_keys(result, namespace='some_namespace', name='some_content',
                version=None, scm=None, src='some_namespace.some_content')


def test_yaml_parse_just_name_no_namespace_required():
    spec = 'some_content'
    result = parse_spec(spec, resolver=repository_spec_parse.resolve)

    assert_keys(result, name='some_content', version=None, scm=None, src='some_content')


def test_yaml_parse_name_and_version():
    spec = 'some_namespace.some_content,1.0.0'
    result = parse_spec(spec)

    assert_keys(result, namespace='some_namespace', name='some_content',
                version='1.0.0', scm=None, src='some_namespace.some_content')


def test_yaml_parse_name_and_version_key_value():
    spec = 'some_namespace.some_content,version=1.0.0'
    result = parse_spec(spec)

    assert_keys(result, namespace='some_namespace', name='some_content',
                version='1.0.0', scm=None, src='some_namespace.some_content')


def test_yaml_parse_name_github_url():
    spec = 'git+https://github.com/geerlingguy/ansible-role-awx.git,1.0.0'
    result = parse_spec(spec)

    assert_keys(result, name='ansible-role-awx', version='1.0.0', scm='git', src='https://github.com/geerlingguy/ansible-role-awx.git')


def test_yaml_parse_name_github_url_keyword_version():
    spec = 'git+https://github.com/geerlingguy/ansible-role-awx.git,version=1.0.0'
    result = parse_spec(spec)

    assert_keys(result, name='ansible-role-awx', version='1.0.0', scm='git', src='https://github.com/geerlingguy/ansible-role-awx.git')


def test_yaml_parse_name_non_github_url():
    buf = 'git+https://git.example.com/geerlingguy/ansible-role-awx.git,1.0.0'
    result = parse_spec(buf)

    assert_keys(result, name='ansible-role-awx', version='1.0.0', scm='git', src='https://git.example.com/geerlingguy/ansible-role-awx.git')


# See https://github.com/ansible/galaxy-cli/wiki/Content-Versioning#versions-in-galaxy-cli
#
# "When providing a version, provide the semantic version with or without the leading 'v' or 'V'."
#
def test_yaml_parse_name_and_version_leading_V():
    spec = 'some_namespace.some_content,V1.0.0'
    result = parse_spec(spec)

    assert_keys(result, namespace='some_namespace', name='some_content',
                version='V1.0.0', scm=None, src='some_namespace.some_content')


def test_yaml_parse_name_and_version_leading_v():
    spec = 'some_namespace.some_content,v1.0.0'
    result = parse_spec(spec)

    assert_keys(result, namespace='some_namespace', name='some_content',
                version='v1.0.0', scm=None, src='some_namespace.some_content')


def test_yaml_parse_name_and_version_leading_v_no_namespace_required():
    spec = 'some_content,v1.0.0'
    result = parse_spec(spec, resolver=repository_spec_parse.resolve)

    assert_keys(result, name='some_content',
                version='v1.0.0', scm=None, src='some_content')


# proving a name and a version as comma separated key values
def test_yaml_parse_name_with_name_key_value():
    spec = 'some_content,name=other_name'
    result = parse_spec(spec, resolver=repository_spec_parse.resolve)

    # TODO: what should 'src' be for this cases?
    assert_keys(result, name='other_name', version=None, scm=None, src='some_content')


# def test_yaml_parse_a_list_of_strings():
#    spec = ['some_content', 'something_else']
#    parse_spec(spec)

    # TODO: verify we get the right exception


def test_yaml_parse_an_empty_dict():
    spec = {}
    result = parse_spec(spec)

    assert_keys(result, name=None, version=None, scm=None, src=None)


def test_yaml_parse_a_dict():
    spec = {'name': 'some_name',
            'version': '1.2.4',
            'src': 'galaxy.role'}
    result = parse_spec(spec)

    assert_keys(result, namespace='galaxy', name='some_name',
                version='1.2.4', scm=None, src='galaxy.role')


def test_yaml_parse_a_dict_with_extra_invalid_keys():
    spec = {'name': 'some_name',
            'version': '1.2.4',
            'src': 'galaxy.role',
            'favorite_tea': 'sweet',
            'module': 'some_module',
            'url': 'http://example.com/galaxy/foo'}
    result = parse_spec(spec)

    assert_keys(result, namespace='galaxy', name='some_name',
                version='1.2.4', scm=None, src='galaxy.role')
    result_keys = set(result.keys())
    valid_keys = set(VALID_ROLE_SPEC_KEYS)
    valid_keys.add('namespace')
    valid_keys.add('fetch_method')
    extra_keys = result_keys.difference(valid_keys)
    assert not extra_keys, \
        'Found extra invalid keys in the result. extra_keys=%s, result=%s, valid_keys=%s' % \
        (extra_keys, result, valid_keys)

    assert_keys(result, namespace='galaxy', name='some_name',
                version='1.2.4', scm=None, src='galaxy.role')


# comments in yaml_parse.py indicate src='galaxy.role,version,name' is supposed to work
def test_yaml_parse_a_dict_with_conflicts():
    spec = {'name': 'some_name1',
            'version': '1.2.3',
            'src': 'galaxy.role,1.0.0,some_name2'}
    result = parse_spec(spec)

    assert_keys(result, namespace='galaxy', name='some_name1',
                version='1.2.3', scm=None, src='galaxy.role,1.0.0,some_name2')


def test_yaml_parse_a_old_style_role_dict():
    spec = {'role': 'some_role',
            'version': '1.2.3',
            'src': 'galaxy.role'}
    result = parse_spec(spec)

    # FIXME: requiring a different set of asserts to test the results for the role
    #        case points to this method doing too many things, being passed too many different things,
    #        and returning too many different things
    name = 'some_role'
    src = 'galaxy.role'
    assert isinstance(result, dict)
    assert result['name'] == name, \
        'repository_spec name=%s does not match expected name=%s' % (result['name'], name)
    assert result['version'] == '1.2.3'
    assert result['src'] == src, \
        'repository_spec src=%s does not match expected src=%s' % (result['src'], src)


# FIXME: I'm not real sure what the result of this is supposed to be
def test_yaml_parse_a_comma_sep_style_role_dict_with_version():
    src = 'galaxy.role,1.2.3'
    spec = {'src': src}
    result = parse_spec(spec, resolver=galaxy_repository_spec.resolve)

    # FIXME: wtf is 'src' expected to look like here?
    assert_keys(result, namespace='galaxy', name='role', version='1.2.3',
                scm=None, src='galaxy.role,1.2.3')


# FIXME: I'm not real sure what the result of this is supposed to be
def test_yaml_parse_a_comma_sep_style_role_dict_with_name_version():
    src = 'galaxy.role,1.2.3,some_role'
    spec = {'src': src}
    result = parse_spec(spec, resolver=galaxy_repository_spec.resolve)

    # FIXME: wtf is 'src' expected to look like here?
    assert_keys(result, namespace='galaxy', name='some_role', version='1.2.3', scm=None, src=src)
